# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

cdef extern from "Python.h":
    object PyBytes_FromStringAndSize(char*,Py_ssize_t)
    char* PyBytes_AsString(object)

cdef inline object allocate_memory(int size, char ** i):
    if size < 0:
        size = 0
    cdef object ob = PyBytes_FromStringAndSize(NULL, size)
    i[0] = PyBytes_AsString(ob)
    return ob

cdef extern from "common_c.h":
    cdef cppclass Vector:
        float x, y, z
        Vector()
        Vector(float x, float y, float z)
        void set(float x, float y, float z)

    struct LongVector:
        int x, y, z

    void destroy_vector(Vector*)

cdef inline int check_default_int(int value, int default) except -1:
    if value != default:
        raise ValueError(
            'check_default() failed: was %s, should be %s' % (value, default))
    return 0

cdef class Quaternion
cdef class Vertex3:
    cdef Vector * value
    cdef bint is_ref
    cpdef Quaternion get_rotation_to(self, Vertex3 A)

cdef class Quaternion:
    cdef public:
        double w, x, y, z
    cpdef Quaternion set_angle_axis(self, double radians, Vertex3 axis)
    cpdef Quaternion slerp(self, Quaternion q, double t)
    cpdef Quaternion nlerp(self, Quaternion q, double t)
    cpdef Vertex3 transform_vector(self, Vertex3 v)
    cpdef Vertex3 inverse_transform_vector(self, Vertex3 v)

cdef inline Vertex3 create_proxy_vector(Vector * v):
    cdef Vertex3 new_vector = Vertex3(is_ref = True)
    new_vector.value = v
    return new_vector

cdef inline Vertex3 create_vertex3(float x, float y, float z):
    # faster way of creating Vertex3 instances
    cdef Vertex3 new_vertex = Vertex3(is_ref = True)
    new_vertex.value = new Vector(x, y, z)
    new_vertex.is_ref = False
    return new_vertex



"""
Pinpoint players and bookmark tunnels without making the enemy aware.
Markers attempt to help teams communicate and so, cooperate.

Markers are just symbols built with blocks at the very top of the map, which
only your team can see. They'll show up on your teammates' map and minimap,
and can be brought up in a number of ways.

* Double-tapping the SNEAK key lets you quickly spot an enemy's location.
* When your intel is stolen a marker is dropped where it just was, shadowing
  its position and eliminating "I forgot where it was" situations.
* Enemy positions are automatically revealed when the intel is captured,
  mimicking the old 'radar' reward.
* Say !tunnel in teamchat to make an arrow where you're standing, instruct other
  players to do something for a change with !build, or meet up with someone
  else using !here.

Players can type /clear to get rid of markers on their own screen.
Admins can disable or enable placing of new markers using /togglemarkers.
/togglemarkers <player> toggles only that player's ability to do so.

Any functionality can be disabled switching off SHADOW_INTEL, REVEAL_ENEMIES,
VV_ENABLED and CHAT_MARKERS below.

Commands
^^^^^^^^

* ``/clear`` clears all markers *admin only*
* ``/togglemarkers <player>`` toggles a player's ability to place markers *admin only*
* ``/markers`` gives instructions

.. codeauthor:: hompy
"""

import csv
from io import StringIO
from collections import deque, defaultdict
from functools import partial
from itertools import islice, chain
from random import choice
from twisted.internet.reactor import callLater, seconds
from pyspades.world import cube_line
from pyspades.contained import BlockAction, BlockLine, SetColor, ChatMessage
from pyspades.common import make_color, to_coordinates
from pyspades.constants import *
from piqueserver.commands import command, admin, get_player, player_only

SHADOW_INTEL = True  # if True, shows where the intel used to be before taken
REVEAL_ENEMIES = True  # if True, mimics old intel reveal behavior when captured
VV_ENABLED = True  # if True, enables spotting by pressing SNEAK two times
CHAT_MARKERS = True  # if True, enables !build !tunnel etc teamchat triggers

S_SPOTTED = '(enemy spotted at {coords}!)'
S_CLEARED = 'Markers cleared'
S_FAIL = "Couldn't place the remote marker because you were looking at the sky"
S_WAIT = 'You must wait until you can place another marker'
S_PLAYER_ENABLED = '{player} can place markers again'
S_PLAYER_DISABLED = '{player} is disallowed from placing markers'
S_ENABLED = 'Markers have been enabled'
S_DISABLED = 'Markers have been disabled'
S_TEAMCHAT = 'Markers work only IN TEAMCHAT! Type /markers for more info'
S_REACHED_LIMIT = 'Your team has too many markers of that kind!'
S_HELP = [
    "* Press V twice to SPOT an enemy, showing it on your teammates' map",
    '* You can also say "!help" IN TEAMCHAT to drop a marker where you are',
    '* (other markers: !build !tunnel !here)'
]

COOLDOWN = 10.0
VV_TIMEFRAME = 0.5
THERE_RAY_LENGTH = 192.0
ENEMY_EXPIRE_DISTANCE_SQUARED = 18.0 ** 2


@command(admin_only=True)
@player_only
def clear(connection):
    connection.destroy_markers()
    return S_CLEARED


@command('togglemarkers', admin_only=True)
def toggle_markers(connection, player=None):
    protocol = connection.protocol
    if player is not None:
        player = get_player(protocol, player)
        player.allow_markers = not player.allow_markers
        message = S_PLAYER_ENABLED if player.allow_markers else S_PLAYER_DISABLED
        message = message.format(player=player.name)
        protocol.send_chat(message, irc=True)
    else:
        protocol.allow_markers = not protocol.allow_markers
        message = S_ENABLED if protocol.allow_markers else S_DISABLED
        connection.protocol.send_chat(message, irc=True)


@command()
@player_only
def markers(connection):
    connection.send_lines(S_HELP)


class BaseMarker():
    name = 'Marker'
    triggers = []
    background = None
    background_class = None
    duration = None
    color = None
    random_colors = None
    team_color = False
    lines = []
    points = []
    always_there = False
    maximum_instances = None
    expire_call = None
    z = 0

    def __init__(self, protocol, team, x, y):
        self.protocol = protocol
        self.team = team
        self.x = x
        self.y = y
        if self.random_colors:
            self.color = choice(self.random_colors)
        elif self.team_color:
            self.color = make_color(*team.color)
        self.blocks = set()
        base_lines, base_points = self.lines, self.points
        self.lines, self.points = [], []
        for line in base_lines:
            self.make_line(*line)
        for point in base_points:
            self.make_block(*point)
        # find markers we're colliding with
        has_timer = self.duration is not None
        collisions = []
        current_time = seconds()
        worst_time = current_time + self.duration if has_timer else None
        for marker in protocol.markers:
            intersect = marker.blocks & self.blocks
            if intersect:
                self.blocks -= intersect
                collisions.append(marker)
                if has_timer and marker.expire_call:
                    worst_time = min(worst_time, marker.expire_call.getTime())
        # forward expiration time so that colliding markers vanish all at once
        if has_timer:
            delay = worst_time - current_time
            self.expire_call = callLater(delay, self.expire)
        self.build()
        team.marker_count[self.__class__] += 1
        protocol.markers.append(self)
        if self.background_class:
            self.background = self.background_class(protocol, team, x, y)

    def release(self):
        if self.expire_call and self.expire_call.active():
            self.expire_call.cancel()
        self.expire_call = None
        self.team.marker_count[self.__class__] -= 1
        self.protocol.markers.remove(self)

    def expire(self):
        self.destroy()
        self.release()
        if self.background:
            self.background.expire()

    @classmethod
    def is_triggered(cls, chat):
        return any(word in chat for word in cls.triggers)

    def make_block(self, x, y):
        x += self.x
        y += self.y
        if x < 0 or y < 0 or x >= 512 or y >= 512:
            return
        block = (x, y, self.z)
        self.blocks.add(block)
        self.points.append(block)

    def make_line(self, x1, y1, x2, y2):
        x1 = max(0, min(511, self.x + x1))
        y1 = max(0, min(511, self.y + y1))
        x2 = max(0, min(511, self.x + x2))
        y2 = max(0, min(511, self.y + y2))
        z = self.z
        line = (x1, y1, z, x2, y2, z)
        self.blocks.update(cube_line(*line))
        self.lines.append(line)

    def build(self, sender=None):
        sender = sender or self.protocol.broadcast_contained
        self.send_color(sender)
        for line in self.lines:
            self.send_line(sender, *line)
        for point in self.points:
            self.send_block(sender, *point)

    def destroy(self, sender=None):
        # breaking a single block would make it come tumbling down, so we have
        # to destroy them all at once
        sender = sender or self.protocol.broadcast_contained
        for block in self.blocks:
            self.send_block_remove(sender, *block)

    def send_color(self, sender):
        set_color = SetColor()
        set_color.value = self.color
        set_color.player_id = 32
        sender(set_color, team=self.team)

    def send_block(self, sender, x, y, z, value=BUILD_BLOCK):
        block_action = BlockAction()
        block_action.value = value
        block_action.player_id = 32
        block_action.x = x
        block_action.y = y
        block_action.z = z
        sender(block_action, team=self.team)

    def send_line(self, sender, x1, y1, z1, x2, y2, z2):
        block_line = BlockLine()
        block_line.player_id = 32
        block_line.x1 = x1
        block_line.y1 = y1
        block_line.z1 = z1
        block_line.x2 = x2
        block_line.y2 = y2
        block_line.z2 = z2
        sender(block_line, team=self.team)

    def send_block_remove(self, sender, x, y, z):
        self.send_block(sender, x, y, z, DESTROY_BLOCK)


def parse_string_map(xs_and_dots):
    """greedily attempt to get the least amount of lines and blocks required
    to build the shape. best (worst) function ever"""

    # format to List[str] with spaces removed, e.g. ["...XX.XX", "..X.XX.."]
    rows = [s.replace(" ", "") for s in xs_and_dots.splitlines() if s.strip()]
    lines, points = [], []
    if not rows:
        return lines, points

    width, height = len(rows[0]), len(rows)

    # offset of the top left to the center
    off_x, off_y = -width // 2, -height // 2

    for y, row in enumerate(rows):
        # if anyone understands this code, tell us
        columns = [''.join(l[y:]).split('.', 1)[0] for l in zip(*rows)]
        it = enumerate(columns)
        for x, column in it:
            h = len(row[x:].split('.', 1)[0])
            v = len(column)
            if h == v == 0:
                continue
            if max(h, v) == 1:
                points.append((x + off_x, y + off_y))
            elif h >= v:
                lines.append((x + off_x, y + off_y, x +
                              off_x + h - 1, y + off_y))
                row = '.' * (x + h) + row[(x + h):]
                next(islice(it, h, h), None)  # forward the iterator
            else:
                lines.append((x + off_x, y + off_y, x +
                              off_x, y + off_y + v - 1))
                rows[y:y + v] = (r[:x] + '.' + r[x + 1:]
                                 for r in rows[y:y + v])
    return lines, points


class EnemyBackground(BaseMarker):
    color = make_color(0, 0, 0)
    s = """
    . . X X X X X . .
    . X X X X X X X .
    X X X X X X X X X
    X X X X X X X X X
    X X X X X X X X X
    X X X X X X X X X
    X X X X X X X X X
    . X X X X X X X .
    . . X X X X X . .
    """


class Enemy(BaseMarker):
    name = 'Point'
    background_class = EnemyBackground
    duration = 40.0
    always_there = True
    color = make_color(255, 0, 0)
    s = """
    . . X X X . .
    . X X X X X .
    X X X X X X X
    X X X X X X X
    X X X X X X X
    . X X X X X .
    . . X X X . .
    """


class Here(BaseMarker):
    name = 'Circle'
    triggers = ['!here']
    duration = 60.0
    random_colors = [
        make_color(192, 255, 0),  # lime green
        make_color(255, 255, 0),  # yellow
        make_color(255, 192, 0),  # orange
        make_color(255, 192, 255),  # light pink
        make_color(0, 192, 255)  # light blue
    ]
    s = """
    . . X X X X X . .
    . X X X X X X X .
    X X X . . . X X X
    X X . . . . . X X
    X X . . . . . X X
    X X . . . . . X X
    X X X . . . X X X
    . X X X X X X X .
    . . X X X X X . .
    """


class BackupBackground(BaseMarker):
    color = make_color(0, 0, 0)
    s = """
    . X X X X .
    X X X X X X
    X X X X X X
    X X X X X X
    X X X X X X
    X X X X X X
    X X X X X X
    . X X X X .
    . X X X X .
    . X X X X .
    . X X X X .
    . X X X X .
    . X X X X .
    """


class Backup(BaseMarker):
    name = 'Exclamation'
    background_class = BackupBackground
    triggers = ['!backup', '!support', '!airstrike', '!danger', '!help']
    duration = 2 * 60.0
    color = make_color(255, 0, 0)
    s = """
    . X X .
    X X X X
    X X X X
    X X X X
    X X X X
    . X X .
    . X X .
    . X X .
    . . . .
    . X X .
    . X X .
    """


class Intel(BaseMarker):
    name = 'Intel'
    color = make_color(255, 255, 255)
    s = """
    . . . X X X X X X . . .
    . . . . . . . . . . . .
    . . . . X X X X . . . .
    X . . . . . . . . . . X
    X . X . . X X . . X . X
    X . X . X X X X . X . X
    X . X . X X X X . X . X
    X . X . . X X . . X . X
    X . . . . . . . . . . X
    . . . . X X X X . . . .
    . . . . . . . . . . . .
    . . . X X X X X X . . .
    """


class BuildBackground(BaseMarker):
    color = make_color(255, 255, 255)
    s = """
    . . . . . X X X . . . . .
    . . . X X X X X X X . . .
    . X X X X X X X X X X X .
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    X X X X X X X X X X X X X
    . . X X X X X X X X X . .
    . . . . X X X X X . . . .
    """


class Build(BaseMarker):
    name = 'Cube'
    triggers = ['!build', '!bunker', '!bridge', '!fort']
    duration = 5 * 60.0
    maximum_instances = 6
    team_color = True
    background_class = BuildBackground
    s = """
    . . . . . X . . . . .
    . . . X X X X X . . .
    . X X X X X X X X X .
    X . . X X X X X . . X
    X X X . . X . . X X X
    X X X X X . X X X X X
    X X X X X . X X X X X
    X X X X X . X X X X X
    X X X X X . X X X X X
    X X X X X . X X X X X
    . . X X X . X X X . .
    . . . . X . X . . . .
    """


class TunnelBackground(BaseMarker):
    color = make_color(255, 255, 255)
    s = """
    . . X X X X X X X . .
    . . X X X X X X X . .
    . . X X X X X X X . .
    . . X X X X X X X . .
    X X X X X X X X X X X
    X X X X X X X X X X X
    X X X X X X X X X X X
    . X X X X X X X X X .
    . . X X X X X X X . .
    . . . X X X X X . . .
    . . . . X X X . . . .
    """


class Tunnel(BaseMarker):
    name = 'Arrow'
    background_class = TunnelBackground
    triggers = ['!tunnel', '!arrow']
    duration = 5 * 60.0
    maximum_instances = 6
    team_color = True
    s = """
    . . X X X X X . .
    . . X X X X X . .
    . . X X X X X . .
    . . X X X X X . .
    X X X X X X X X X
    . X X X X X X X .
    . . X X X X X . .
    . . . X X X . . .
    . . . . X . . . .
    """


class NumberMarker(BaseMarker):
    duration = Enemy.duration
    background_class = Enemy
    color = make_color(255, 255, 255)
    always_there = True


class Zero(NumberMarker):
    name = '0'
    triggers = ['!0']
    s = """
    . X X X .
    X X . X X
    X X . X X
    X X . X X
    X X . X X
    X X . X X
    . X X X .
    """


class One(NumberMarker):
    name = '1'
    triggers = ['!1']
    s = """
    . X X X .
    . . X X .
    . . X X .
    . . X X .
    . . X X .
    . . X X .
    . X X X X
    """


class Two(NumberMarker):
    name = '2'
    triggers = ['!2']
    s = """
    . X X X .
    X X . X X
    . . . X X
    . . X X .
    . X X . .
    X X . . .
    X X X X X
    """


class Three(NumberMarker):
    name = '3'
    triggers = ['!3']
    s = """
    . X X X .
    X X . X X
    . . . X X
    . . X X .
    . . . X X
    X X . X X
    . X X X .
    """


class Four(NumberMarker):
    name = '4'
    triggers = ['!4']
    s = """
    X X . X X
    X X . X X
    X X . X X
    . X X X X
    . . . X X
    . . . X X
    . . . X X
    """


class Five(NumberMarker):
    name = '5'
    triggers = ['!5']
    s = """
    X X X X X
    X X . . .
    X X . . .
    X X X X .
    . . . X X
    . . . X X
    X X X X .
    """


class Six(NumberMarker):
    name = '6'
    triggers = ['!6']
    s = """
    . X X X .
    X X . X X
    X X . . .
    X X X X .
    X X . X X
    X X . X X
    . X X X .
    """


class Seven(NumberMarker):
    name = '7'
    triggers = ['!7']
    s = """
    X X X X X
    . . . X X
    . . . X X
    . . X X .
    . X X . .
    X X . . .
    X X . . .
    """


class Eight(NumberMarker):
    name = '8'
    triggers = ['!8']
    s = """
    . X X X .
    X X . X X
    X X . X X
    . X X X .
    X X . X X
    X X . X X
    . X X X .
    """


class Nine(NumberMarker):
    name = '9'
    triggers = ['!9']
    s = """
    . X X X .
    X X . X X
    X X . X X
    . X X X X
    . . . X X
    X X . X X
    . X X X .
    """


# in triggering order
number_markers = [Zero, One, Two, Three, Four, Five, Six, Seven, Eight, Nine]
trigger_markers = [Tunnel, Backup, Build] + number_markers + [Here]
other_markers = [Enemy, Intel]
background_markers = []

# turn bitmap into line and block instructions
for cls in chain(trigger_markers, other_markers, background_markers):
    # FIXME: This code does not actually work! It only places the background,
    # not the foreground
    if cls.background_class:
        # This modifies background-markers while looping over it, which is very
        # bad practice. However in this case it is accidentally fine.
        background_markers.append(cls.background_class)
    cls.lines, cls.points = parse_string_map(cls.s)


def apply_script(protocol, connection, config):
    class MarkerConnection(connection):
        allow_markers = True
        last_marker = None
        sneak_presses = None

        def send_markers(self):
            def is_self(player): return player is self
            send_me = partial(self.protocol.broadcast_contained, rule=is_self)
            for marker in self.protocol.markers:
                marker.build(send_me)

        def destroy_markers(self):
            def is_self(player): return player is self
            send_me = partial(self.protocol.broadcast_contained, rule=is_self)
            for marker in self.protocol.markers:
                marker.destroy(send_me)

        def make_marker(self, marker_class, location):
            marker_max = marker_class.maximum_instances
            if (marker_max is not None and
                    self.team.marker_count[marker_class] >= marker_max):
                self.send_chat(S_REACHED_LIMIT)
                return
            new_marker = marker_class(self.protocol, self.team, *location)
            self.last_marker = seconds()

        def on_animation_update(self, jump, crouch, sneak, sprint):
            markers_allowed = (VV_ENABLED and self.allow_markers and
                               self.protocol.allow_markers)
            if markers_allowed and sneak and self.world_object.sneak != sneak:
                now = seconds()
                if self.last_marker is None or now - self.last_marker > COOLDOWN:
                    presses = self.sneak_presses
                    presses.append(now)
                    if len(presses) == 2 and presses[0] >= now - VV_TIMEFRAME:
                        location = self.get_there_location()
                        if location:
                            coords = to_coordinates(*location)
                            chat_message = ChatMessage()
                            chat_message.chat_type = CHAT_TEAM
                            chat_message.player_id = self.player_id
                            chat_message.value = S_SPOTTED.format(
                                coords=coords)
                            self.protocol.broadcast_contained(
                                chat_message, team=self.team)
                            self.make_marker(Enemy, location)
                            presses.clear()
            return connection.on_animation_update(self, jump, crouch, sneak,
                                                  sprint)

        def on_chat(self, value, global_message):
            markers_allowed = self.allow_markers and self.protocol.allow_markers
            if CHAT_MARKERS and markers_allowed and not self.team.spectator:
                chat = value.lower()
                try:
                    marker_class = next(cls for cls in trigger_markers if
                                        cls.is_triggered(chat))
                except StopIteration:
                    pass
                else:
                    if global_message:
                        self.send_chat(S_TEAMCHAT)
                    elif (self.last_marker is not None and
                          seconds() - self.last_marker <= COOLDOWN):
                        self.send_chat(S_WAIT)
                    else:
                        location = None
                        if marker_class.always_there or 'there' in chat:
                            location = self.get_there_location()
                            if location is None:
                                self.send_chat(S_FAIL)
                        else:
                            x, y, z = self.get_location()
                            location = (x + 6 if self.team.id ==
                                        0 else x - 6, y)
                        if location:
                            self.make_marker(marker_class, location)
            return connection.on_chat(self, value, global_message)

        def on_login(self, name):
            self.send_markers()
            self.sneak_presses = deque(maxlen=2)
            connection.on_login(self, name)

        def on_team_changed(self, old_team):
            if old_team and not old_team.spectator:
                new_team, self.team = self.team, old_team
                self.destroy_markers()
                self.team = new_team
            if self.team and not self.team.spectator:
                self.send_markers()
            connection.on_team_changed(self, old_team)

        def on_kill(self, killer, type, grenade):
            x1, y1, z1 = self.get_location()
            closest, best_distance = None, None
            for marker in self.protocol.markers:
                if marker.name == Enemy.name:
                    x, y = x1 - marker.x, y1 - marker.y
                    distance = x * x + y * y
                    if distance > ENEMY_EXPIRE_DISTANCE_SQUARED:
                        continue
                    if not closest or distance < best_distance:
                        closest, best_distance = marker, distance
            if closest:
                closest.expire()
            return connection.on_kill(self, killer, type, grenade)

        def on_flag_take(self):
            if SHADOW_INTEL and self.protocol.allow_markers:
                enemy_team = self.team.other
                x, y, z = self.get_location()
                marker = Intel(self.protocol, enemy_team, x, y)
                enemy_team.intel_marker = marker
            return connection.on_flag_take(self)

        def on_flag_capture(self):
            enemy_team = self.team.other
            if SHADOW_INTEL and enemy_team.intel_marker:
                enemy_team.intel_marker.expire()
                enemy_team.intel_marker = None
            if REVEAL_ENEMIES and self.protocol.allow_markers:
                delay = 0.25
                for call in self.team.marker_calls:
                    if call.active():
                        call.cancel()
                self.team.marker_calls = []
                for player in enemy_team.get_players():
                    x, y, z = player.get_location()
                    delay += 0.15
                    call = callLater(
                        delay, Enemy, self.protocol, self.team, x, y)
                    self.team.marker_calls.append(call)
            connection.on_flag_capture(self)

        def on_flag_drop(self):
            enemy_team = self.team.other
            if SHADOW_INTEL and enemy_team.intel_marker:
                enemy_team.intel_marker.expire()
                enemy_team.intel_marker = None
            connection.on_flag_drop(self)

        def get_there_location(self):
            location = self.world_object.cast_ray(THERE_RAY_LENGTH)
            return location[:2] if location else None

    class MarkerProtocol(protocol):
        allow_markers = True
        markers = None

        def on_map_change(self, map):
            for team in (self.blue_team, self.green_team):
                team.intel_marker = None
                team.marker_calls = []
                team.marker_count = defaultdict(int)
            self.markers = []
            protocol.on_map_change(self, map)

        def on_map_leave(self):
            for marker in self.markers[:]:
                marker.release()
            self.markers = None
            for team in (self.blue_team, self.green_team):
                team.intel_marker = None
                for call in team.marker_calls:
                    if call.active():
                        call.cancel()
                team.marker_calls = None
                team.marker_count = None
            protocol.on_map_leave(self)

    return MarkerProtocol, MarkerConnection



DEF INT_ERROR = -0xFFFFFFFF >> 1
DEF LONG_LONG_ERROR = -0xFFFFFFFFFFFFFFFF >> 1
DEF FLOAT_ERROR = float('nan')

cdef extern from "<sstream>" namespace "std":
    cdef cppclass stringstream:
        pass

cdef class ByteReader:
    cdef char * data
    cdef char * pos
    cdef char * end
    cdef int start, size
    cdef object input

    cdef char * check_available(self, int size) except NULL
    cpdef read(self, int bytecount = ?)
    cpdef int readByte(self, bint unsigned = ?) except INT_ERROR
    cpdef int readShort(self, bint unsigned = ?, bint big_endian = ?) \
                        except INT_ERROR
    cpdef long long readInt(self, bint unsigned = ?, bint big_endian = ?) \
                            except LONG_LONG_ERROR
    cpdef float readFloat(self, bint big_endian = ?) except? FLOAT_ERROR
    cpdef bytes readString(self, int size = ?)
    cpdef ByteReader readReader(self, int size = ?)
    cpdef int dataLeft(self)
    cdef void _skip(self, int bytecount)
    cpdef skipBytes(self, int bytecount)
    cpdef rewind(self, int value)
    cpdef seek(self, size_t pos)
    cpdef size_t tell(self)

cdef class ByteWriter:
    cdef stringstream * stream

    cdef void writeSize(self, char * data, int size)
    cpdef write(self, data)
    cpdef writeByte(self, int value, bint unsigned = ?)
    cpdef writeShort(self, int value, bint unsigned = ?,
                     bint big_endian = ?)
    cpdef writeInt(self, long long value, bint unsigned = ?,
                   bint big_endian = ?)
    cpdef writeFloat(self, float value, bint big_endian = ?)
    cpdef writeStringSize(self, char * value, int size)
    cpdef writeString(self, value, int size = ?)
    cpdef pad(self, int bytecount)
    cpdef rewind(self, int bytecount)
    cpdef size_t tell(self)



# Copyright (c) Mathias Kaerlev 2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from weakref import WeakSet
from twisted.internet import reactor
from twisted.internet.task import LoopingCall


class Scheduler:

    def __init__(self, protocol):
        self.protocol = protocol
        self.calls = WeakSet()
        self.loops = WeakSet()

    def call_later(self, *arg, **kw):
        call = reactor.callLater(*arg, **kw)
        self.calls.add(call)
        return call

    def call_end(self, *arg, **kw):
        call = self.protocol.call_end(*arg, **kw)
        self.calls.add(call)
        return call

    def loop_call(self, delay, func, *arg, **kw):
        loop = LoopingCall(func, *arg, **kw)
        loop.start(delay, False)
        self.loops.add(loop)
        return loop

    def reset(self):
        for call in self.calls:
            if call.active():
                call.cancel()
        for loop in self.loops:
            if loop.running:
                loop.stop()
        self.calls = WeakSet()
        self.loops = WeakSet()



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

"""
pyspades - Ace of Spades network protocol implementations
"""

__author__ = 'Mathias Kaerlev'
__copyright__ = 'Copyright 2011, Mathias Kaerlev'
__credits__ = ['Mathias "Mat^2" Kaerlev']
__version__ = '0.1'
__email__ = 'matpow2@gmail.com'

__all__ = ['vxl', 'bytes', 'contained', 'collision', 'common', 'constants',
           'loaders', 'master', 'packet', 'protocol', 'server', 'tools',
           'types']



"""
Greets specified people entering with messages

Options
^^^^^^^

.. code-block:: guess

   [welcome]
   welcomes = { nota = "Hi notafile", feik = "Hi feik" }

.. codeauthor: mat^2
"""
from piqueserver.config import config

welcome_config = config.section("welcome")
welcomes_option = welcome_config.option("welcomes", {})


def apply_script(protocol, connection, config):
    welcomes = welcomes_option.get()

    class EnterConnection(connection):

        def on_login(self, name):
            if name in welcomes:
                self.protocol.send_chat(welcomes[name])
            connection.on_login(self, name)
    return protocol, EnterConnection



"""
Match script, useful for public matches. Features verbose announcements
on IRC and a custom timer.

Commands
^^^^^^^^

* ``/timer`` starts timer *admin only*
* ``/stoptimer`` stops timer *admin only*
* ``/startrecord`` starts a fresh records *admin only*
* ``/stoprecord`` clears the current record *admin only*
* ``/saverecord`` save record to a file *admin only*

.. codeauthor: mat^2
"""
import os

from twisted.internet import reactor
from twisted.internet.task import LoopingCall

import json
from piqueserver.commands import command, admin
from piqueserver.config import config


config_dir = config.config_dir


@command('timer', admin_only=True)
def start_timer(connection, end):
    return connection.protocol.start_timer(int(end) * 60)


@command('stoptimer', admin_only=True)
def stop_timer(connection, end):
    return connection.protocol.stop_timer()


@command('startrecord', admin_only=True)
def start_record(connection):
    connection.protocol.start_record()
    return 'Recording started.'


@command('stoprecord', admin_only=True)
def stop_record(connection):
    connection.protocol.stop_record()
    return 'Recording stopped.'


@command('saverecord', admin_only=True)
def save_record(connection, name):
    if not connection.protocol.save_record(name):
        return 'No record file available.'
    return 'Record saved.'


def apply_script(protocol, connection, config):
    class MatchConnection(connection):

        def on_flag_take(self):
            self.add_message(
                "%s took %s's flag!" %
                (self.printable_name, self.team.other.name.lower()))
            return connection.on_flag_take(self)

        def on_flag_drop(self):
            self.add_message(
                "%s dropped %s's flag!" %
                (self.printable_name, self.team.other.name.lower()))
            return connection.on_flag_drop(self)

        def on_flag_capture(self):
            self.add_message(
                "%s captured %s's flag!" %
                (self.printable_name, self.team.other.name.lower()))
            return connection.on_flag_capture(self)

        def on_kill(self, killer, type, grenade):
            if killer is None:
                killer = self
            self.add_message("%s was killed by %s!" %
                             (self.printable_name, killer.printable_name))
            self.protocol.add_kill(self, killer)
            return connection.on_kill(self, killer, type, grenade)

        def add_message(self, value):
            self.protocol.messages.append(value)

    class MatchProtocol(protocol):
        timer_left = None
        timer_call = None
        timer_end = None
        record = None

        def __init__(self, *arg, **kw):
            protocol.__init__(self, *arg, **kw)
            self.messages = []
            self.send_message_loop = LoopingCall(self.display_messages)
            self.send_message_loop.start(3)

        def start_timer(self, end):
            if self.timer_end is not None:
                return 'Timer is running already.'
            self.timer_end = reactor.seconds() + end
            self.send_chat('Timer started, ending in %s minutes' % (end / 60),
                           irc=True)
            self.display_timer(True)

        def stop_timer(self):
            if self.timer_call is not None:
                self.timer_call.cancel()
                self.send_chat('Timer stopped.')
                self.timer_call = None
            else:
                return 'No timer in progress.'

        def display_timer(self, silent=False):
            time_left = self.timer_end - reactor.seconds()
            minutes_left = time_left / 60.0
            next_call = 60
            if not silent:
                if time_left <= 0:
                    self.send_chat('Timer ended!', irc=True)
                    self.timer_end = None
                    return
                elif minutes_left <= 1:
                    self.send_chat('%s seconds left' % int(time_left),
                                   irc=True)
                    next_call = max(1, int(time_left / 2.0))
                else:
                    self.send_chat('%s minutes left' % int(minutes_left),
                                   irc=True)
            self.timer_call = reactor.callLater(next_call, self.display_timer)

        def display_messages(self):
            if not self.messages:
                return
            message = self.messages.pop(0)
            self.irc_say(message)

        # recording

        def add_kill(self, player, killing_player):
            if self.record is None:
                return
            self.get_record(player.name)['deaths'] += 1
            self.get_record(killing_player.name)['kills'] += 1

        def get_record(self, name):
            try:
                return self.record[name]
            except KeyError:
                record = {'deaths': 0, 'kills': 0}
                self.record[name] = record
                return record

        def start_record(self):
            self.record = {}

        def stop_record(self):
            self.record = None

        def save_record(self, name):
            if self.record is None:
                return False
            path = os.path.join(config_dir, "record_{}.json".format(name))
            with open(path, "w") as f:
                json.dump(self.record, f)
            return True

    return MatchProtocol, MatchConnection



import collections
import math
import random
import re
import shlex
import textwrap
from itertools import product
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import enet
from twisted.internet import reactor
from twisted.logger import Logger

from pyspades import contained as loaders
from pyspades import world
from pyspades.collision import collision_3d, vector_collision
from pyspades.common import Vertex3, get_color, make_color
from pyspades.constants import *
from pyspades.constants import (BLOCK_TOOL, CTF_MODE, ERROR_FULL,
                                ERROR_TOO_MANY_CONNECTIONS,
                                ERROR_WRONG_VERSION, FALL_KILL, HEAD,
                                HEADSHOT_KILL, HIT_TOLERANCE,
                                MAX_BLOCK_DISTANCE, MAX_POSITION_RATE, MELEE,
                                MELEE_DISTANCE, MELEE_KILL,
                                RAPID_WINDOW_ENTRIES, SPADE_TOOL,
                                TC_CAPTURE_DISTANCE, TC_MODE, WEAPON_KILL,
                                WEAPON_TOOL)
from pyspades.mapgenerator import ProgressiveMapGenerator
from pyspades.packet import call_packet_handler, register_packet_handler
from pyspades.protocol import BaseConnection
from pyspades.team import Team
from pyspades.weapon import WEAPONS

log = Logger()


tc_data = loaders.TCState()

def check_nan(*values) -> bool:
    for value in values:
        if math.isnan(value):
            return True
    return False


def parse_command(value: str) -> Tuple[str, Sequence[str]]:
    try:
        splitted = shlex.split(value)
    except ValueError:
        # shlex failed. let's just split per space
        splitted = value.split(' ')
    if splitted:
        command = splitted.pop(0)
    else:
        command = ''
    return command, splitted


class SlidingWindow:
    def __init__(self, entries: Any) -> None:
        self.entries = entries
        self.window = collections.deque()  # type: Deque

    def add(self, value) -> None:
        self.window.append(value)
        if len(self.window) <= self.entries:
            return
        self.window.popleft()

    def check(self) -> bool:
        return len(self.window) == self.entries

    def get(self) -> Any:
        return self.window[0], self.window[-1]


class ServerConnection(BaseConnection):
    address = None  # Tuple[int, int]
    player_id = None
    map_packets_sent = 0
    team = None  # type: Team
    weapon = None
    weapon_object = None
    name = None
    kills = 0
    hp = None
    tool = None
    color = (0x70, 0x70, 0x70)
    grenades = None
    blocks = None
    spawn_call = None
    respawn_time = None
    saved_loaders = None
    last_refill = None
    last_block_destroy = None
    filter_visibility_data = False
    filter_animation_data = False
    freeze_animation = False
    filter_weapon_input = False
    speedhack_detect = False
    rubberband_distance = 10
    rapid_hack_detect = False
    timers = None
    world_object = None  # type: world.Character
    last_block = None
    map_data = None
    last_position_update = None
    local = False

    def __init__(self, *arg, **kw) -> None:
        BaseConnection.__init__(self, *arg, **kw)
        protocol = self.protocol
        address = self.peer.address
        self.total_blocks_removed = 0
        self.address = (address.host, address.port)
        self.respawn_time = protocol.respawn_time
        self.rapids = SlidingWindow(RAPID_WINDOW_ENTRIES)
        self.client_info = {}
        self.proto_extensions = {}  # type: Dict[int, int]
        self.line_build_start_pos = None

    def on_connect(self) -> None:
        if self.local:
            return
        if self.peer.eventData != self.protocol.version:
            log.debug("{player} kicked: wrong protocol version {version}",
                      player=self, version=self.peer.eventData)
            self.disconnect(ERROR_WRONG_VERSION)
            return
        max_players = min(32, self.protocol.max_players)
        if len(self.protocol.connections) > max_players:
            self.disconnect(ERROR_FULL)
            return
        if self.protocol.max_connections_per_ip:
            shared = [conn for conn in
                      self.protocol.connections.values()
                      if conn.address[0] == self.address[0]]
            if len(shared) > self.protocol.max_connections_per_ip:
                self.disconnect(ERROR_TOO_MANY_CONNECTIONS)
                return
        if not self.disconnected:
            log.debug("sending map data to {player}", player=self)
            self._connection_ack()

    def loader_received(self, loader: enet.Packet) -> None:
        """
        called when a loader i.e. packet is received.
        calls the packet handler registered with
        @register_packet_handler
        """
        if self.player_id is None:
            return
        call_packet_handler(self, loader)

    @register_packet_handler(loaders.ProtocolExtensionInfo)
    def on_ext_info_received(self, contained: loaders.ProtocolExtensionInfo) -> None:
        self.proto_extensions = dict(contained.extensions)
        log.debug("received extinfo {extinfo} from {player}",
                  extinfo=self.proto_extensions,
                  player=self)

    @register_packet_handler(loaders.ExistingPlayer)
    @register_packet_handler(loaders.ShortPlayerData)
    def on_new_player_recieved(self, contained: loaders.ExistingPlayer) -> None:
        if self.team is not None and not self.team.spectator:
            # This player has already joined the game as a full player.
            # Existingplayer may only be sent if in the limbo or spectator
            # modes. Without this check, they could respawn themselves
            # instantly on any team they wanted.
            log.debug("{} tried sending an ExistingPlayer packet while not in"
                      " limbo or spectator mode".format(self))
            return

        old_team = self.team
        team = self.protocol.teams[contained.team]
        log.debug("{user} wants to join {team}",
                  user=self, team=team, teamid=contained.team)

        ret = self.on_team_join(team)
        if ret is False:
            team = self.protocol.team_spectator
        elif ret is not None:
            team = ret

        self.team = team
        if self.name is None:
            name = contained.name
            self.name = self.protocol.get_name(name)
            self.protocol.players[self.player_id] = self
            self.on_login(self.name)
        else:
            self.on_team_changed(old_team)
        self.set_weapon(contained.weapon, True)
        if self.protocol.speedhack_detect and not self.local:
            self.speedhack_detect = True
        if self.protocol.rubberband_distance is not None:
            self.rubberband_distance = self.protocol.rubberband_distance
        if not self.local:
            self.rapid_hack_detect = True
        if team.spectator:
            if self.world_object is not None:
                self.world_object.delete()
                self.world_object = None
        # send kill packets for dead players
        for player in self.protocol.players.values():
            if (player.player_id != self.player_id and player.world_object
                    and player.world_object.dead):
                kill_action = loaders.KillAction()
                kill_action.killer_id = player.player_id
                kill_action.player_id = player.player_id
                kill_action.kill_type = FALL_KILL
                self.send_contained(kill_action)

        self.spawn()

    @register_packet_handler(loaders.OrientationData)
    def on_orientation_update_recieved(self, contained: loaders.OrientationData) -> None:
        if not self.hp:
            return
        x, y, z = contained.x, contained.y, contained.z
        if check_nan(x, y, z):
            self.on_hack_attempt(
                'Invalid orientation data received')
            return
        returned = self.on_orientation_update(x, y, z)
        if returned == False:
            return
        if returned is not None:
            x, y, z = returned
        self.world_object.set_orientation(x, y, z)

    @register_packet_handler(loaders.PositionData)
    def on_position_update_recieved(self, contained: loaders.PositionData) -> None:
        if not self.hp:
            return
        current_time = reactor.seconds()
        last_update = self.last_position_update
        self.last_position_update = current_time
        if last_update is not None:
            dt = current_time - last_update
            if dt < MAX_POSITION_RATE:
                self.set_location()
                return
        x, y, z = contained.x, contained.y, contained.z
        if check_nan(x, y, z):
            self.on_hack_attempt(
                'Invalid position data received')
            return
        if not self.check_speedhack(x, y, z):
            # vanilla behaviour
            self.set_location()
            return
        if not self.freeze_animation:
            self.world_object.set_position(x, y, z)
            self.on_position_update()
        if self.filter_visibility_data:
            return
        game_mode = self.protocol.game_mode
        if game_mode == CTF_MODE:
            other_flag = self.team.other.flag
            if vector_collision(self.world_object.position,
                                self.team.base):
                if other_flag.player is self:
                    self.capture_flag()
                self.check_refill()
            if other_flag.player is None and vector_collision(
                    self.world_object.position, other_flag):
                self.take_flag()
        elif game_mode == TC_MODE:
            for entity in self.protocol.entities:
                collides = vector_collision(
                    entity, self.world_object.position, TC_CAPTURE_DISTANCE)
                if self in entity.players:
                    if not collides:
                        entity.remove_player(self)
                else:
                    if collides:
                        entity.add_player(self)
                if collides and vector_collision(entity,
                                                 self.world_object.position):
                    self.check_refill()

    @register_packet_handler(loaders.WeaponInput)
    def on_weapon_input_recieved(self, contained: loaders.WeaponInput) -> None:
        if not self.hp:
            return
        primary = contained.primary
        secondary = contained.secondary
        if self.world_object.primary_fire != primary:
            # player has pressed or released primary fire
            if self.tool == WEAPON_TOOL:
                self.weapon_object.set_shoot(primary)
            if self.tool == WEAPON_TOOL or self.tool == SPADE_TOOL:
                self.on_shoot_set(primary)

        if self.world_object.secondary_fire != secondary:
            # player has pressed or released secondary fire
            self.on_secondary_fire_set(secondary)

            if secondary and self.tool == BLOCK_TOOL:
                # hook into here to save the start location of the line build.
                # This is needed so we can check if the player was actually at
                # the starting location of the line build when it was started.
                # this is inspired by 1AmYF's fbpatch2.py script
                position = self.world_object.position
                self.line_build_start_pos = position.copy()
                self.on_line_build_start()

        # remember the current state of the mouse buttons
        self.world_object.primary_fire = primary
        self.world_object.secondary_fire = secondary
        if self.filter_weapon_input:
            return
        contained.player_id = self.player_id
        self.protocol.send_contained(contained, sender=self)

    @register_packet_handler(loaders.InputData)
    def on_input_data_recieved(self, contained: loaders.InputData) -> None:
        if not self.hp:
            return
        world_object = self.world_object
        returned = self.on_walk_update(contained.up, contained.down,
                                       contained.left, contained.right)
        if returned is not None:
            up, down, left, right = returned
            if (up != contained.up or down != contained.down or
                    left != contained.left or right != contained.right):
                (contained.up, contained.down, contained.left,
                    contained.right) = returned
                # XXX unsupported
                # self.send_contained(contained)
        if not self.freeze_animation:
            world_object.set_walk(contained.up, contained.down,
                                  contained.left, contained.right)
        contained.player_id = self.player_id
        z_vel = world_object.velocity.z
        if contained.jump and not (z_vel >= 0 and z_vel < 0.017):
            contained.jump = False
        # XXX unsupported for now
        # returned = self.on_animation_update(contained.primary_fire,
            # contained.secondary_fire, contained.jump,
            # contained.crouch)
        # if returned is not None:
            # fire1, fire2, jump, crouch = returned
            # if (fire1 != contained.primary_fire or
            # fire2 != contained.secondary_fire or
            # jump != contained.jump or
            # crouch != contained.crouch):
            # (contained.primary_fire, contained.secondary_fire,
            # contained.jump, contained.crouch) = returned
            # self.send_contained(contained)
        returned = self.on_animation_update(
            contained.jump, contained.crouch, contained.sneak,
            contained.sprint)
        if returned is not None:
            jump, crouch, sneak, sprint = returned
            if (jump != contained.jump or crouch != contained.crouch or
                    sneak != contained.sneak or sprint != contained.sprint):
                (contained.jump, contained.crouch, contained.sneak,
                    contained.sprint) = returned
                self.send_contained(contained)
        if not self.freeze_animation:
            world_object.set_animation(
                contained.jump, contained.crouch, contained.sneak,
                contained.sprint)
        if self.filter_visibility_data or self.filter_animation_data:
            return
        self.protocol.send_contained(contained, sender=self)

    @register_packet_handler(loaders.WeaponReload)
    def on_reload_recieved(self, contained) -> None:
        if not self.hp:
            return
        self.weapon_object.reload()
        if self.filter_animation_data:
            return
        contained.player_id = self.player_id
        self.protocol.send_contained(contained, sender=self)

    @register_packet_handler(loaders.HitPacket)
    def on_hit_recieved(self, contained):
        if not self.hp:
            return
        world_object = self.world_object
        value = contained.value
        is_melee = value == MELEE
        if not is_melee and self.weapon_object.is_empty():
            return
        try:
            player = self.protocol.players[contained.player_id]
        except KeyError:
            return
        valid_hit = world_object.validate_hit(player.world_object,
                                              value, HIT_TOLERANCE)
        if not valid_hit:
            return
        position1 = world_object.position
        position2 = player.world_object.position
        if is_melee:
            if not vector_collision(position1, position2,
                                    MELEE_DISTANCE):
                return
            hit_amount = self.protocol.melee_damage
        else:
            hit_amount = self.weapon_object.get_damage(
                value, position1, position2)
        if is_melee:
            kill_type = MELEE_KILL
        elif contained.value == HEAD:
            kill_type = HEADSHOT_KILL
        else:
            kill_type = WEAPON_KILL
        returned = self.on_hit(hit_amount, player, kill_type, None)
        if returned == False:
            return
        elif returned is not None:
            hit_amount = returned
        player.hit(hit_amount, self, kill_type)

    @register_packet_handler(loaders.GrenadePacket)
    def on_grenade_recieved(self, contained: loaders.GrenadePacket) -> None:
        if not self.hp:
            return
        if check_nan(contained.value, *contained.position) or check_nan(*contained.velocity):
            self.on_hack_attempt("Invalid grenade data")
            return
        if not self.grenades:
            return
        self.grenades -= 1
        if not self.check_speedhack(*contained.position):
            contained.position = self.world_object.position.get()
        if self.on_grenade(contained.value) == False:
            return
        grenade = self.protocol.world.create_object(
            world.Grenade, contained.value,
            Vertex3(*contained.position), None,
            Vertex3(*contained.velocity), self.grenade_exploded)
        grenade.team = self.team
        log.debug("{player!r} ({world_object!r}) created {grenade!r}",
                  grenade=grenade, world_object=self.world_object, player=self)
        self.on_grenade_thrown(grenade)
        if self.filter_visibility_data:
            return
        contained.player_id = self.player_id
        self.protocol.send_contained(contained,
                                     sender=self)

    @register_packet_handler(loaders.SetTool)
    def on_tool_change_recieved(self, contained: loaders.SetTool) -> None:
        if not self.hp:
            return
        if self.on_tool_set_attempt(contained.value) == False:
            return
        old_tool = self.tool
        self.tool = contained.value
        if old_tool == WEAPON_TOOL:
            self.weapon_object.set_shoot(False)
        if self.tool == WEAPON_TOOL:
            self.on_shoot_set(self.world_object.primary_fire)
            self.weapon_object.set_shoot(
                self.world_object.primary_fire)
        self.world_object.set_weapon(self.tool == WEAPON_TOOL)
        self.on_tool_changed(self.tool)
        if self.filter_visibility_data or self.filter_animation_data:
            return
        set_tool = loaders.SetTool()
        set_tool.player_id = self.player_id
        set_tool.value = contained.value
        self.protocol.send_contained(set_tool, sender=self, save=True)

    @register_packet_handler(loaders.SetColor)
    def on_color_change_recieved(self, contained: loaders.SetColor) -> None:
        if not self.hp:
            return
        color = get_color(contained.value)
        if self.on_color_set_attempt(color) == False:
            return
        self.color = color
        self.on_color_set(color)
        if self.filter_animation_data:
            return
        contained.player_id = self.player_id
        self.protocol.send_contained(contained, sender=self,
                                     save=True)

    @register_packet_handler(loaders.BlockAction)
    def on_block_action_recieved(self, contained: loaders.BlockAction) -> None:
        world_object = self.world_object
        if not self.hp:
            return
        value = contained.value
        if value == BUILD_BLOCK:
            interval = TOOL_INTERVAL[BLOCK_TOOL]
        elif self.tool == WEAPON_TOOL:
            if self.weapon_object.is_empty():
                return
            interval = WEAPON_INTERVAL[self.weapon]
        else:
            interval = TOOL_INTERVAL[self.tool]
        current_time = reactor.seconds()
        last_time = self.last_block
        self.last_block = current_time
        if (self.rapid_hack_detect and last_time is not None and
                current_time - last_time < interval):
            self.rapids.add(current_time)
            if self.rapids.check():
                start, end = self.rapids.get()
                if end - start < MAX_RAPID_SPEED:
                    log.info('RAPID HACK: {window}', window=self.rapids.window)
                    self.on_hack_attempt('Rapid hack detected')
            return
        map = self.protocol.map
        x = contained.x
        y = contained.y
        z = contained.z
        if z >= 62:
            return
        if value == BUILD_BLOCK:
            self.blocks -= 1
            pos = world_object.position
            if self.blocks < -BUILD_TOLERANCE:
                return
            elif not collision_3d(pos.x, pos.y, pos.z, x, y, z,
                                  MAX_BLOCK_DISTANCE):
                return
            elif self.on_block_build_attempt(x, y, z) == False:
                return
            elif not map.build_point(x, y, z, self.color):
                return
            self.on_block_build(x, y, z)
        else:
            if not map.get_solid(x, y, z):
                return
            pos = world_object.position
            if self.tool == SPADE_TOOL and not collision_3d(
                    pos.x, pos.y, pos.z, x, y, z, MAX_DIG_DISTANCE):
                return
            if self.on_block_destroy(x, y, z, value) == False:
                return
            elif value == DESTROY_BLOCK:
                count = map.destroy_point(x, y, z)
                if count:
                    self.total_blocks_removed += count
                    self.blocks = min(50, self.blocks + 1)
                    self.on_block_removed(x, y, z)
            elif value == SPADE_DESTROY:
                for xyz in ((x, y, z), (x, y, z + 1), (x, y, z - 1)):
                    count = map.destroy_point(*xyz)
                    if count:
                        self.total_blocks_removed += count
                        self.on_block_removed(*xyz)
            self.last_block_destroy = reactor.seconds()
        block_action = loaders.BlockAction()
        block_action.x = x
        block_action.y = y
        block_action.z = z
        block_action.value = contained.value
        block_action.player_id = self.player_id
        self.protocol.send_contained(block_action, save=True)
        self.protocol.update_entities()

    @register_packet_handler(loaders.BlockLine)
    def on_block_line_recieved(self, contained):
        if not self.hp:
            return  # dead players can't build

        map_ = self.protocol.map

        x1, y1, z1 = (contained.x1, contained.y1, contained.z1)
        x2, y2, z2 = (contained.x2, contained.y2, contained.z2)
        pos = self.world_object.position
        start_pos = self.line_build_start_pos

        if (not map_.is_valid_position(x1, y1, z1)
                or not map_.is_valid_position(x2, y2, z2)):
            return  # coordinates are out of bounds

        # ensure that the player is currently within tolerance of the location
        # that the line build ended at
        if not collision_3d(pos.x, pos.y, pos.z, x2, y2, z2,
                            MAX_BLOCK_DISTANCE):
            return

        # ensure that the player was within tolerance of the location
        # that the line build started at
        if not collision_3d(start_pos.x, start_pos.y, start_pos.z, x1, y1, z1,
                            MAX_BLOCK_DISTANCE):
            return

        points = world.cube_line(x1, y1, z1, x2, y2, z2)

        if not points:
            return

        if len(points) > (self.blocks + BUILD_TOLERANCE):
            return

        if self.on_line_build_attempt(points) is False:
            return

        for point in points:
            x, y, z = point
            if map_.get_solid(x, y, z):
                continue
            if not map_.build_point(x, y, z, self.color):
                break

        self.blocks -= len(points)
        self.on_line_build(points)
        contained.player_id = self.player_id
        self.protocol.send_contained(contained, save=True)
        self.protocol.update_entities()

    @register_packet_handler(loaders.ChatMessage)
    def on_chat_message_recieved(self, contained: loaders.ChatMessage) -> None:
        if not self.name:
            return
        value = contained.value
        if value.startswith('/'):
            self.on_command(*parse_command(value[1:]))
        else:
            global_message = contained.chat_type == CHAT_ALL
            result = self.on_chat(value, global_message)
            if result == False:
                return
            elif result is not None:
                value = result
            contained.chat_type = CHAT_ALL if global_message else CHAT_TEAM
            contained.value = value
            contained.player_id = self.player_id
            if global_message:
                team = None
            else:
                team = self.team
            for player in self.protocol.players.values():
                if not player.deaf:
                    if team is None or team is player.team:
                        player.send_contained(contained)
            self.on_chat_sent(value, global_message)

    @register_packet_handler(loaders.FogColor)
    def on_fog_color_recieved(self, contained):
        # FIXME: this theoretically might anyone to set the fog...
        # do we even need this?
        if not self.name:
            return
        color = get_color(contained.color)
        self.on_command('fog', [str(item) for item in color])

    @register_packet_handler(loaders.ChangeWeapon)
    def on_weapon_change_recieved(self, contained):
        if not self.name:
            return
        if self.on_weapon_set(contained.weapon) == False:
            return
        self.weapon = contained.weapon
        self.set_weapon(self.weapon)

    @register_packet_handler(loaders.ChangeTeam)
    def on_team_change_recieved(self, contained):
        if not self.name:
            return
        team = self.protocol.teams[contained.team]
        ret = self.on_team_join(team)
        if ret is False:
            return
        team = ret or team
        self.set_team(team)

    @register_packet_handler(loaders.HandShakeReturn)
    def on_handshake_recieved(self, contained: loaders.HandShakeReturn) -> None:
        version_request = loaders.VersionRequest()
        self.protocol.send_contained(version_request)

    @register_packet_handler(loaders.VersionResponse)
    def on_version_info_recieved(self, contained: loaders.VersionResponse) -> None:
        self.client_info["version"] = contained.version
        self.client_info["os_info"] = contained.os_info
        # TODO: Make this a dict lookup instead
        if contained.client == 'o':
            self.client_info["client"] = "OpenSpades"
        elif contained.client == 'B':
            self.client_info["client"] = "BetterSpades"
            # BetterSpades currently sends the client name in the OS info to
            # deal with old scripts that don't recognize the 'B' indentifier
            match = re.match(r"\ABetterSpades \((.*)\)\Z", contained.os_info)
            if match:
                self.client_info["os_info"] = match.groups()[0]
        elif contained.client == 'a':
            self.client_info["client"] = "ACE"
        else:
            self.client_info["client"] = "Unknown({})".format(contained.client)

        # send extension info to clients that support this packet.
        # skip openspades <= 0.1.3 https://github.com/piqueserver/piqueserver/issues/504
        if contained.client == 'o' and contained.version <= (0, 1, 3):
            log.debug("not sending version request to OpenSpades <= 0.1.3")
        else:
            ext_info = loaders.ProtocolExtensionInfo()
            ext_info.extensions = []
            self.send_contained(ext_info)

    @property
    def client_string(self):
        client = self.client_info.get("client", "Unknown")
        os = self.client_info.get("os_info", "Unknown")
        version = self.client_info.get("version", None)
        version_string = "Unknown" if version is None else ".".join(
            map(str, version))
        if client == os == version_string == "Unknown":
            client = "Probably Voxlap"
            os = "Windows"
            version_string = "0.75"
        return "{} v{} on {}".format(client, version_string, os)

    def check_speedhack(self, x: float, y: float, z: float, distance: None = None) -> bool:
        if not self.speedhack_detect:
            return True
        if distance is None:
            distance = self.rubberband_distance
        position = self.world_object.position
        return (math.fabs(x - position.x) < distance and
                math.fabs(y - position.y) < distance and
                math.fabs(z - position.z) < distance)

    # backwards compatability
    is_valid_position = check_speedhack

    def check_refill(self):
        last_refill = self.last_refill
        if (last_refill is None or
                reactor.seconds() - last_refill > self.protocol.refill_interval):
            self.last_refill = reactor.seconds()
            if self.on_refill() != False:
                self.refill()

    def get_location(self):
        position = self.world_object.position
        return position.x, position.y, position.z

    def is_location_free(self, x, y, z):
        return (self.protocol.map.get_solid(x, y, z) == 0 and
                self.protocol.map.get_solid(x, y, z + 1) == 0 and
                self.protocol.map.get_solid(x, y, z + 2) == 0 and
                self.protocol.map.get_solid(x, y, z + 3) == 1)

    def set_location_safe(self, location, center=True):
        x, y, z = location

        if center:
            x -= 0.5
            y -= 0.5
            z += 0.5

        x = int(x)
        y = int(y)
        z = int(z)

        # search for valid locations near the specified point
        for pos in self.protocol.pos_table:
            if self.is_location_free(x + pos[0], y + pos[1], z + pos[2]):
                self.set_location((x + pos[0], y + pos[1], z + pos[2]))
                return True
        return False

    def set_location(self, location=None):
        if location is None:
            # used for rubberbanding
            position = self.world_object.position
            x, y, z = position.x, position.y, position.z
        else:
            x, y, z = location
            if self.world_object is not None:
                self.world_object.set_position(x, y, z)
            x += 0.5
            y += 0.5
            z -= 0.5
        position_data = loaders.PositionData()
        position_data.x = x
        position_data.y = y
        position_data.z = z
        self.send_contained(position_data)

    def refill(self, local: bool = False) -> None:
        self.hp = 100
        self.grenades = 3
        self.blocks = 50
        self.weapon_object.restock()
        if not local:
            restock = loaders.Restock()
            self.send_contained(restock)

    def respawn(self) -> None:
        if self.spawn_call is None:
            self.spawn_call = reactor.callLater(
                self.get_respawn_time(), self.spawn)

    def get_spawn_location(self) -> Tuple[int, int, int]:
        game_mode = self.protocol.game_mode
        if game_mode == TC_MODE:
            try:
                base = random.choice(list(self.team.get_entities()))
                return base.get_spawn_location()
            except IndexError:
                pass
        return self.team.get_random_location(True)

    def get_respawn_time(self) -> float:
        if not self.respawn_time:
            return 0
        if self.protocol.respawn_waves:
            offset = reactor.seconds() % self.respawn_time
        else:
            offset = 0
        return self.respawn_time - offset

    def spawn(self, pos: None = None) -> None:
        self.spawn_call = None
        if self.team is None:
            return
        spectator = self.team.spectator
        create_player = loaders.CreatePlayer()
        if not spectator:
            if pos is None:
                x, y, z = self.get_spawn_location()
                x += 0.5
                y += 0.5
                z -= 2.4
            else:
                x, y, z = pos
            returned = self.on_spawn_location((x, y, z))
            if returned is not None:
                x, y, z = returned
            if self.world_object is not None:
                self.world_object.set_position(x, y, z, True)
            else:
                position = Vertex3(x, y, z)
                self.world_object = self.protocol.world.create_object(
                    world.Character, position, None, self._on_fall)
            self.world_object.dead = False
            self.tool = WEAPON_TOOL
            self.refill(True)
            create_player.x = x
            create_player.y = y
            create_player.z = z
            create_player.weapon = self.weapon
        create_player.player_id = self.player_id
        create_player.name = self.name
        create_player.team = self.team.id
        if self.filter_visibility_data and not spectator:
            self.send_contained(create_player)
        else:
            self.protocol.send_contained(create_player, save=True)
        if not spectator:
            self.on_spawn((x, y, z))

        if not self.client_info:
            handshake_init = loaders.HandShakeInit()
            self.send_contained(handshake_init)

    def take_flag(self):
        if not self.hp:
            return
        flag = self.team.other.flag
        if flag.player is not None:
            return
        if self.on_flag_take() == False:
            return
        flag.player = self
        intel_pickup = loaders.IntelPickup()
        intel_pickup.player_id = self.player_id
        self.protocol.send_contained(intel_pickup, save=True)

    def capture_flag(self):
        other_team = self.team.other
        flag = other_team.flag
        player = flag.player
        if player is not self:
            return
        self.add_score(10)  # 10 points for intel
        self.team.score += 1
        self.on_flag_capture()
        if (self.protocol.max_score not in (0, None) and
                self.team.score >= self.protocol.max_score):
            self.protocol.reset_game(self)
            self.protocol.on_game_end()
        else:
            intel_capture = loaders.IntelCapture()
            intel_capture.player_id = self.player_id
            intel_capture.winning = False
            self.protocol.send_contained(intel_capture, save=True)
            flag = other_team.set_flag()
            flag.update()

    def drop_flag(self) -> None:
        protocol = self.protocol
        game_mode = protocol.game_mode
        if game_mode == CTF_MODE:
            for flag in (protocol.blue_team.flag, protocol.green_team.flag):
                player = flag.player
                if player is not self:
                    continue
                position = self.world_object.position

                # convert to safe coords so the flag can't be dropped out of bounds
                x, y, z = self.protocol.map.get_safe_coords(
                    position.x, position.y, position.z)
                # or inside solid
                z = self.protocol.map.get_z(x, y, z)
                flag.set(x, y, z)

                flag.player = None
                intel_drop = loaders.IntelDrop()
                intel_drop.player_id = self.player_id
                intel_drop.x = flag.x
                intel_drop.y = flag.y
                intel_drop.z = flag.z
                self.protocol.send_contained(intel_drop, save=True)
                self.on_flag_drop()
                break
        elif game_mode == TC_MODE:
            for entity in protocol.entities:
                if self in entity.players:
                    entity.remove_player(self)

    def on_disconnect(self) -> None:
        if self.name is not None:
            self.drop_flag()
            player_left = loaders.PlayerLeft()
            player_left.player_id = self.player_id
            self.protocol.send_contained(player_left, sender=self,
                                         save=True)
            del self.protocol.players[self.player_id]
        if self.player_id is not None:
            self.protocol.player_ids.put_back(self.player_id)
            self.protocol.update_master()
        self.reset()

    def reset(self) -> None:
        if self.spawn_call is not None:
            self.spawn_call.cancel()
            self.spawn_call = None
        if self.world_object is not None:
            self.world_object.delete()
            self.world_object = None
        if self.team is not None:
            old_team = self.team
            self.team = None
            self.on_team_changed(old_team)
        self.on_reset()
        self.name = self.hp = self.world_object = None

    def hit(self, value, by=None, kill_type=WEAPON_KILL):
        if self.hp is None:
            return
        if by is not None and self.team is by.team:
            friendly_fire = self.protocol.friendly_fire
            friendly_fire_on_grief = self.protocol.friendly_fire_on_grief
            if friendly_fire_on_grief:
                if (kill_type == MELEE_KILL and
                        not self.protocol.spade_teamkills_on_grief):
                    return
                hit_time = self.protocol.friendly_fire_time
                if (self.last_block_destroy is None
                        or reactor.seconds() - self.last_block_destroy >= hit_time):
                    return
            elif not friendly_fire:
                return
        self.set_hp(self.hp - value, by, kill_type=kill_type)

    def set_hp(self, value: Union[int, float], hit_by: Optional['ServerConnection'] = None, kill_type: int = WEAPON_KILL,
               hit_indicator: Optional[Tuple[float, float, float]] = None, grenade: Optional[world.Grenade] = None) -> None:
        value = int(value)
        self.hp = max(0, min(100, value))
        if self.hp <= 0:
            self.kill(hit_by, kill_type, grenade)
            return
        set_hp = loaders.SetHP()
        set_hp.hp = self.hp
        set_hp.not_fall = int(kill_type != FALL_KILL)
        if hit_indicator is None:
            if hit_by is not None and hit_by is not self:
                hit_indicator = hit_by.world_object.position.get()
            else:
                hit_indicator = (0, 0, 0)
        x, y, z = hit_indicator
        set_hp.source_x = x
        set_hp.source_y = y
        set_hp.source_z = z
        self.send_contained(set_hp)

    def set_weapon(self, weapon: int, local: bool = False, no_kill: bool = False) -> None:
        self.weapon = weapon
        if self.weapon_object is not None:
            self.weapon_object.reset()
        self.weapon_object = WEAPONS[weapon](self._on_reload)
        if not local:
            change_weapon = loaders.ChangeWeapon()
            self.protocol.send_contained(change_weapon, save=True)
            if not no_kill:
                self.kill(kill_type=CLASS_CHANGE_KILL)

    def set_team(self, team):
        if team is self.team:
            return
        self.drop_flag()
        old_team = self.team
        self.team = team
        self.on_team_changed(old_team)
        if old_team.spectator:
            self.respawn()
        else:
            self.kill(kill_type=TEAM_CHANGE_KILL)

    def kill(self, by: None = None, kill_type: int = WEAPON_KILL, grenade: None = None) -> None:
        if self.hp is None:
            return
        if self.on_kill(by, kill_type, grenade) is False:
            return
        self.drop_flag()
        self.hp = None
        self.weapon_object.reset()
        kill_action = loaders.KillAction()
        kill_action.kill_type = kill_type
        if by is None:
            kill_action.killer_id = kill_action.player_id = self.player_id
        else:
            kill_action.killer_id = by.player_id
            kill_action.player_id = self.player_id
        if by is not None and by is not self:
            by.add_score(1)
        kill_action.respawn_time = self.get_respawn_time() + 1
        self.protocol.send_contained(kill_action, save=True)
        self.world_object.dead = True
        self.respawn()

    def add_score(self, score):
        self.kills += score

    def _connection_ack(self) -> None:
        self._send_connection_data()
        self.send_map(ProgressiveMapGenerator(self.protocol.map))
        if not self.client_info:
            handshake_init = loaders.HandShakeInit()
            self.send_contained(handshake_init)

    def _send_connection_data(self) -> None:
        saved_loaders = self.saved_loaders = []
        if self.player_id is None:
            for player in self.protocol.players.values():
                if player.name is None:
                    continue
                existing_player = loaders.ExistingPlayer()
                existing_player.name = player.name
                existing_player.player_id = player.player_id
                existing_player.tool = player.tool or 0
                existing_player.weapon = player.weapon
                existing_player.kills = player.kills
                existing_player.team = player.team.id
                existing_player.color = make_color(*player.color)
                saved_loaders.append(existing_player.generate())

            self.player_id = self.protocol.player_ids.pop()
            self.protocol.update_master()

        # send initial data
        blue = self.protocol.blue_team
        green = self.protocol.green_team

        state_data = loaders.StateData()
        state_data.player_id = self.player_id
        state_data.fog_color = self.protocol.fog_color
        state_data.team1_color = blue.color
        state_data.team1_name = blue.name
        state_data.team2_color = green.color
        state_data.team2_name = green.name

        game_mode = self.protocol.game_mode

        if game_mode == CTF_MODE:
            blue_base = blue.base
            blue_flag = blue.flag
            green_base = green.base
            green_flag = green.flag
            ctf_data = loaders.CTFState()
            ctf_data.cap_limit = self.protocol.max_score
            ctf_data.team1_score = blue.score
            ctf_data.team2_score = green.score

            ctf_data.team1_base_x = blue_base.x
            ctf_data.team1_base_y = blue_base.y
            ctf_data.team1_base_z = blue_base.z

            ctf_data.team2_base_x = green_base.x
            ctf_data.team2_base_y = green_base.y
            ctf_data.team2_base_z = green_base.z

            if green_flag.player is None:
                ctf_data.team1_has_intel = 0
                ctf_data.team2_flag_x = green_flag.x
                ctf_data.team2_flag_y = green_flag.y
                ctf_data.team2_flag_z = green_flag.z
            else:
                ctf_data.team1_has_intel = 1
                ctf_data.team2_carrier = green_flag.player.player_id

            if blue_flag.player is None:
                ctf_data.team2_has_intel = 0
                ctf_data.team1_flag_x = blue_flag.x
                ctf_data.team1_flag_y = blue_flag.y
                ctf_data.team1_flag_z = blue_flag.z
            else:
                ctf_data.team2_has_intel = 1
                ctf_data.team1_carrier = blue_flag.player.player_id

            state_data.state = ctf_data

        elif game_mode == TC_MODE:
            state_data.state = tc_data

        generated_data = state_data.generate()
        saved_loaders.append(generated_data)

    def grenade_exploded(self, grenade: world.Grenade) -> None:
        if self.name is None or self.team.spectator:
            return
        if grenade.team is not None and grenade.team is not self.team:
            # could happen if the player changed team
            return
        position = grenade.position
        x = position.x
        y = position.y
        z = position.z
        if x < 0 or x > 512 or y < 0 or y > 512 or z < 0 or z > 63:
            return
        x = int(math.floor(x))
        y = int(math.floor(y))
        z = int(math.floor(z))
        for player_list in (self.team.other.get_players(), (self,)):
            for player in player_list:
                if not player.hp:
                    continue
                damage = grenade.get_damage(player.world_object.position)
                if damage == 0:
                    continue
                returned = self.on_hit(damage, player, GRENADE_KILL, grenade)
                if returned == False:
                    continue
                elif returned is not None:
                    damage = returned
                player.set_hp(player.hp - damage, self,
                              hit_indicator=position.get(), kill_type=GRENADE_KILL,
                              grenade=grenade)
        if self.on_block_destroy(x, y, z, GRENADE_DESTROY) == False:
            return
        map = self.protocol.map
        for n_x, n_y, n_z in product(range(x - 1, x + 2), range(y - 1, y + 2), range(z - 1, z + 2)):
            count = map.destroy_point(n_x, n_y, n_z)
            if count:
                self.total_blocks_removed += count
                self.on_block_removed(n_x, n_y, n_z)
        block_action = loaders.BlockAction()
        block_action.x = x
        block_action.y = y
        block_action.z = z
        block_action.value = GRENADE_DESTROY
        block_action.player_id = self.player_id
        self.protocol.send_contained(block_action, save=True)
        self.protocol.update_entities()

    def _on_fall(self, damage: int) -> None:
        if not self.hp:
            return
        returned = self.on_fall(damage)
        if returned is False:
            return
        elif returned is not None:
            damage = returned
        self.set_hp(self.hp - damage, kill_type=FALL_KILL)

    def _on_reload(self):
        weapon_reload = loaders.WeaponReload()
        weapon_reload.player_id = self.player_id
        weapon_reload.clip_ammo = self.weapon_object.current_ammo
        weapon_reload.reserve_ammo = self.weapon_object.current_stock
        self.send_contained(weapon_reload)

    def send_map(self, data: Optional[ProgressiveMapGenerator] = None) -> None:
        if data is not None:
            self.map_data = data
            map_start = loaders.MapStart()
            map_start.size = data.get_size()
            self.send_contained(map_start)
        elif self.map_data is None:
            return

        if not self.map_data.data_left():
            log.debug("done sending map data to {player}", player=self)
            self.map_data = None
            for data in self.saved_loaders:
                packet = enet.Packet(bytes(data), enet.PACKET_FLAG_RELIABLE)
                self.peer.send(0, packet)
            self.saved_loaders = None
            self.on_join()
            return
        for _ in range(10):
            if not self.map_data.data_left():
                break
            map_data = loaders.MapChunk()
            map_data.data = self.map_data.read(8192)
            self.send_contained(map_data)

    def continue_map_transfer(self) -> None:
        self.send_map()

    def send_data(self, data):
        self.protocol.transport.write(data, self.address)

    def send_chat(self, value: str, global_message: bool = False) -> None:
        if self.deaf:
            return
        chat_message = loaders.ChatMessage()
        if not global_message:
            chat_message.chat_type = CHAT_SYSTEM
            prefix = ''
        else:
            chat_message.chat_type = CHAT_TEAM
            # 34 is guaranteed to be out of range!
            chat_message.player_id = 35
            prefix = self.protocol.server_prefix + ' '

        lines = textwrap.wrap(value, MAX_CHAT_SIZE - len(prefix) - 1)

        for line in lines:
            chat_message.value = '{}{}'.format(prefix, line)
            self.send_contained(chat_message)

    def send_chat_warning(self, message):
        """
        Send a warning message. This gets displayed as a yellow popup
        with sound for OpenSpades clients
        """
        self.send_chat("%% " + str(message))

    def send_chat_notice(self, message):
        """
        Send a notice. This gets displayed as a popup for OpenSpades
        clients
        """
        self.send_chat("N% " + str(message))

    def send_chat_error(self, message):
        """
        Send a error message. This gets displayed as a red popup with
        sound for OpenSpades clients
        """
        self.send_chat("!% " + str(message))

    def send_chat_status(self, message):
        """
        Send a status message. This gets displayed in the center of the
        screen for OpenSpades clients
        """
        self.send_chat("C% " + str(message))

    # events/hooks

    def on_join(self):
        pass

    def on_login(self, name):
        pass

    def on_spawn(self, pos: Tuple[float, float, float]) -> None:
        pass

    def on_spawn_location(self, pos: Tuple[float, float, float]) -> None:
        pass

    def on_chat(self, value, global_message):
        pass

    def on_chat_sent(self, value: str, global_message: bool) -> None:
        pass

    def on_command(self, command, parameters):
        pass

    def on_hit(self, hit_amount, hit_player, kill_type, grenade):
        pass

    def on_kill(self, killer, kill_type, grenade):
        pass

    def on_team_join(self, team):
        pass

    def on_team_changed(self, old_team: Team) -> None:
        pass

    def on_tool_set_attempt(self, tool: int) -> None:
        pass

    def on_tool_changed(self, tool: int) -> None:
        pass

    def on_grenade(self, time_left):
        pass

    def on_grenade_thrown(self, grenade: world.Grenade) -> None:
        pass

    def on_block_build_attempt(self, x, y, z):
        pass

    def on_block_build(self, x, y, z):
        pass

    def on_line_build_start(self):
        """called when the player has pressed the mouse button to start
        line-building"""

    def on_line_build_attempt(self, points):
        pass

    def on_line_build(self, points):
        pass

    def on_block_destroy(self, x, y, z, mode):
        pass

    def on_block_removed(self, x, y, z):
        pass

    def on_refill(self):
        pass

    def on_color_set_attempt(self, color: Tuple[int, int, int]) -> None:
        pass

    def on_color_set(self, color: Tuple[int, int, int]) -> None:
        pass

    def on_flag_take(self):
        pass

    def on_flag_capture(self):
        pass

    def on_flag_drop(self):
        pass

    def on_hack_attempt(self, reason):
        pass

    def on_position_update(self) -> None:
        pass

    def on_weapon_set(self, value):
        pass

    def on_fall(self, damage):
        pass

    def on_reset(self):
        pass

    def on_orientation_update(self, x: float, y: float, z: float) -> None:
        pass

    def on_shoot_set(self, fire: int) -> None:
        pass

    def on_secondary_fire_set(self, secondary):
        pass

    def on_walk_update(self, up: bool, down: bool, left: bool, right: bool) -> None:
        pass

    def on_animation_update(self, jump, crouch, sneak, sprint):
        pass

    def __repr__(self):
        return "<{} player_id: {!r}, name: {!r}, address: {!r} at 0x{:x}>".format(
            self.__class__.__name__, self.player_id, self.name, self.address,
            id(self)
        )



# Copyright (c) 2017 Piqueserver development team

# This file is part of piqueserver.

# piqueserver is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# piqueserver is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with piqueserver.  If not, see <http://www.gnu.org/licenses/>.

import collections
import json
import os
import sys

import piqueserver
import toml
from piqueserver.utils import timeparse

# supported config format constants to avoid typos
DEFAULT_FORMAT = 'TOML'
TOML_FORMAT = 'TOML'
JSON_FORMAT = 'JSON'

# global constants we need to know at the start
_path = os.environ.get('XDG_CONFIG_HOME', '~/.config') + '/piqueserver'
DEFAULT_CONFIG_DIR = os.path.expanduser(_path)
MAXMIND_DOWNLOAD = 'https://geolite.maxmind.com/download/geoip/database/GeoLite2-City.tar.gz'
MAXMIND_DOWNLOAD_MD5 = MAXMIND_DOWNLOAD + '.md5'

# (major, minor) versions of python we are supporting
# used on startup to emit a warning if not running on a supported version
SUPPORTED_PYTHONS = ((3, 5), (3, 6), (3, 7), (3, 8))


class ConfigStore():
    '''
    Configuration store that manages global configuration.

    Usage example:

    >>> config = ConfigStore()
    >>> config.load_from_dict({'key1': 'value1'})

    >>> option1 = config.option('key1', default='nothing',
    ...                         validate=lambda x: len(x) > 0)
    >>> section1 = config.section('section1')
    >>> nested_option = section1.option('key1', default=0)

    >>> print(nested_option.get())
    >>> option1.set('hello')

    >>> # underlying dictionary structure will look like
    >>> # {
    >>> #   'key1': 'hello',
    >>> #   'section1': {
    >>> #     'key1': 0
    >>> #   }
    >>> # }
    '''

    def __init__(self):
        self._raw_config = {}
        self._options = {}
        self._sections = {}

        # these are for config that isn't ever stored in the config file (yet)
        self.config_dir = DEFAULT_CONFIG_DIR
        self.config_file = os.path.join(DEFAULT_CONFIG_DIR, "config.toml")

    def _validate_all(self):
        for option in self._options.values():
            option._validate(option.get())
        for section in self._sections.values():
            section._validate_all()

    # https://stackoverflow.com/a/3233356/
    def _nested_update(self, config_dict, updates):
        for k, v in updates.items():
            if isinstance(v, collections.Mapping):
                config_dict[k] = self._nested_update(config_dict.get(k, {}), v)
            else:
                config_dict[k] = v
        return config_dict

    def get_dict(self):
        return self._raw_config

    def load_from_file(self, fobj, format_=DEFAULT_FORMAT):
        '''
        Clear the current configuration and load new configuration from a file-like object
        in a supported format.
        '''
        self._raw_config = {}
        self.update_from_file(fobj, format_)

    def update_from_file(self, fobj, format_=DEFAULT_FORMAT):
        '''
        Updates the configuration from a file-like object. Useful for
        overwriting/updating part of the config without touching the rest.
        '''
        d = {}
        if format_ == TOML_FORMAT:
            d = toml.load(fobj)
        elif format_ == JSON_FORMAT:
            d = json.load(fobj)
        else:
            raise ValueError(
                'Unsupported config file format: {}'.format(format_))
        self.update_from_dict(d)

    def load_from_dict(self, config):
        '''
        Load from a dictionary object directly.
        '''
        self._raw_config = {}
        self.update_from_dict(config)

    def update_from_dict(self, config):
        '''
        Load from a dictionary object directly.
        '''
        self._raw_config = self._nested_update(self._raw_config, config)
        self._validate_all()

    def dump_to_file(self, fobj, format_=DEFAULT_FORMAT):
        '''
        Writes the current configuration to a file-like objection,
        with the format specified by `format_`.
        '''
        if format_ == TOML_FORMAT:
            toml.dump(self._raw_config, fobj)
        elif format_ == JSON_FORMAT:
            json.dump(self._raw_config, fobj, indent=2)
        else:
            raise ValueError(
                'Unsupported config file format: {}'.format(format_))

    def _get(self, name, default=None):
        if name not in self._raw_config:
            self._raw_config[name] = default
        return self._raw_config[name]

    def _set(self, name, value):
        self._raw_config[name] = value

    def check_unused(self):
        '''
        Return the subset of the underlying dictionary that doesn't have any
        corresponding registered options.
        '''

        unused = {}
        for k, v in self.get_dict().items():
            if isinstance(v, collections.Mapping):
                if k in self._sections:
                    section_unused = self._sections[k].check_unused()
                    if section_unused:
                        unused[k] = section_unused
                else:
                    if k not in self._options:
                        unused[k] = v
            else:
                if k not in self._options:
                    unused[k] = v

        return unused

    def option(self, name, default=None, cast=None, validate=None):
        '''
        Register and return a new option object.
        '''
        # TODO: how to handle same option defined twice?
        option = _Option(self, name, default, cast, validate)
        self._options[name] = option
        return option

    def section(self, name):
        '''
        Register and return a new section object.
        '''
        if name in self._sections:
            return self._sections[name]

        section = _Section(self, name)
        self._sections[name] = section
        return section


class _Section(ConfigStore):
    '''
    Represents a section of a configstore. Can be nested arbitarily.
    '''

    def __init__(self, store, name):
        self._store = store
        self._name = name
        self._sections = {}
        self._options = {}

    def get_dict(self):
        return self._store.get_dict().get(self._name, {})

    def load_from_file(self, fobj, format_=DEFAULT_FORMAT):
        raise NotImplementedError()

    def update_from_file(self, fobj, format_=DEFAULT_FORMAT):
        raise NotImplementedError()

    def load_from_dict(self, config):
        self._store._set(self._name, config)

    def update_from_dict(self, config):
        d = self._store._get(self._name, {})
        d.update(config)
        self._store._set(self._name, d)

    def dump_to_file(self, fobj, format_=DEFAULT_FORMAT):
        raise NotImplementedError()

    def _get(self, name, default):
        section = self._store._get(self._name, {})
        if name not in section:
            section[name] = default
            self._store._set(self._name, section)
        return section[name]

    def _set(self, name, value):
        section = self._store._get(self._name, {})
        section[name] = value
        self._store._set(self._name, section)


class _Option():
    '''
    configuration option object, backed by a configuration store
    '''

    def __init__(self, store, name, default, cast, validate):
        '''
        store: a ConfigStore or Section object. Must provide `get()` and `set(value)` methods
        name: the option name (corresponding to the dictionary key for the section/store)
        cast: a transformation function that will be called whenever you retrieve the option's value.
        validate: a function that takes the casted value and should return bool indicating whether it is a valid value
        '''
        self._store = store  # ConfigStore object
        self._name = name
        self._default = default
        self._cast = cast
        self._validate_func = validate

        # get and validate on declaration to make sure all is good
        self._validate(self.get())

    def _validate(self, value):
        '''
        Runs the validation function (if provided) against the value.
        The function should not mutate the value.
        Returns the value if validation function returns a truthy value, otherwise raises an exception
        '''
        if self._validate_func is not None:
            if not self._validate_func(value):
                raise ValueError(
                    'Failed to validate {!r} config option'.format(self._name))
        return value

    def get(self):
        '''
        Returns the option's value at this point in time. Do not rely on this
        to be the same each time it is called, since it will update if the
        configuration is updated/reloaded.
        '''
        value = self._store._get(self._name, self._default)
        if self._cast is not None:
            return self._cast(value)
        return value

    def set(self, value):
        '''
        Sets the option's value. Will first cast and validate.
        '''
        if self._cast is not None:
            value = self._cast(value)
        self._validate(value)
        self._store._set(self._name, value)


# the global instance to be used across all the codebase
config = ConfigStore()


def cast_duration(d) -> int:
    """
    casts duration(1min, 1hr) into seconds.
    If input is an int it returns that unmodified.
    """
    if isinstance(d, int):
        return d
    if not isinstance(d, str):
        raise ValueError("Invalid type")
    seconds = timeparse(d)
    if seconds is None:
        raise ValueError("Invalid duration")
    return seconds



from piqueserver.map import check_rotation, MapNotFound
from piqueserver.commands import command, parse_maps


@command()
def mapname(connection):
    """
    Tell you what's the name of the current map
    /mapname
    """
    return 'Current map: ' + connection.protocol.map_info.name


@command('showrotation')
def show_rotation(connection):
    """
    Tell you the current map rotation
    /showrotation
    """
    return ", ".join(connection.protocol.get_map_rotation())


@command('map', admin_only=True)
def change_planned_map(connection, *pre_maps):
    """
    Set the next map to be loaded after current game ends and inform everyone of it
    /map <mapname>
    """
    name = connection.name
    protocol = connection.protocol

    # parse seed numbering
    maps, _map_list = parse_maps(pre_maps)
    if not maps:
        return 'Invalid map name'

    planned_map = maps[0]
    try:
        protocol.planned_map = check_rotation([planned_map])[0]
        protocol.send_chat('%s changed next map to %s' %
                           (name, planned_map), irc=True)
    except MapNotFound:
        return 'Map %s not found' % (maps[0])


@command('rotation', admin_only=True)
def change_rotation(connection, *pre_maps):
    """
    Change the current map rotation and informs everyone on the server of it
    /rotation <map1> ... <mapN>
    """
    name = connection.name
    protocol = connection.protocol

    maps, map_list = parse_maps(pre_maps)

    if len(maps) == 0:
        return 'Usage: /rotation <map1> <map2> <map3> ...'
    try:
        protocol.set_map_rotation(maps)
    except MapNotFound as e:
        return 'Invalid map in map rotation (%s)' % e.map
    protocol.send_chat("%s changed map rotation to %s." %
                       (name, map_list), irc=True)


@command('rotationadd', admin_only=True)
def rotation_add(connection, *pre_maps):
    """
    Append a given map to the current map rotation and informs everyone on the server of it
    /rotationadd <map>
    """
    name = connection.name

    protocol = connection.protocol

    new_maps, map_list = parse_maps(pre_maps)

    maps = connection.protocol.get_map_rotation()
    map_list = ", ".join(maps) + map_list
    maps.extend(new_maps)

    try:
        protocol.set_map_rotation(maps)
    except MapNotFound as e:
        return 'Invalid map in map rotation (%s)' % e.map
    protocol.send_chat("%s added %s to map rotation." %
                       (name, " ".join(pre_maps)), irc=True)


@command('revertrotation', admin_only=True)
def revert_rotation(connection):
    """
    Invert the current map rotation
    /revertrotation
    """
    protocol = connection.protocol
    name = connection.name
    maps = protocol.config['maps']
    protocol.set_map_rotation(maps)
    protocol.irc_say("* %s reverted map rotation to %s" % (name, maps))


@command('advancemap', admin_only=True)
def advance(connection):
    """
    Force the next map to be immediately loaded instead of waiting for the time limit to end
    /advancemap
    """
    connection.protocol.advance_rotation('Map advance forced.')



"""
Turns your grenades into magical kv6-growing seeds. Or beans!

Use ``/model <filename>`` to load model files. These must be placed inside a
folder named :file:`kv6/` in the config directory. A full path would look like
this: :file:`~/config/kv6/{filename}.kv6`. Then it can be loaded with ``/model
some_model``

Wildcards and subfolders are allowed. Some examples::

    /model building*

Loads all models that start with "building", like "building1" and
"buildingred"::

    /model my_trees/*

Loads all models inside a folder named "my_trees"

When you have many models loaded, each grenade will pick a random one to grow.
To stop littering the map with random objects just type ``/model``.

Pivot
^^^^^

THE PIVOT POINT in kv6 files determines the first block to be placed. The rest
of the model will then follow, growing around it.

The pivot point MUST be sitting on a block, or the model won't load. Checking
'Adjust pivots' in the Tools menu in SLAB6 will show handy coordinates.
To be sure, you can move the pivot to *half* of a block, e.g.: (4.50, 4.50, 8.50)

In a tree kv6, for example, the pivot point would lie on the lowest block of the
tree trunk, so that it grows up and not into the ground.

Works best with 0.75 kv6 models.

You can adjust FLYING_MODELS and GROW_ON_WATER to allow growing in the air and
on water, respectively. These are disabled by default so you can fly high and
sprinkle tree-growing grenades without worrying about unseemly oddities.

Options
^^^^^^^

.. code-block:: guess

    [grownade]

    flying_models = False # if False grenades exploding in midair will be ignored
    grow_on_water = False # if False grenades exploding in water will do nothing

.. codeauthor:: hompy
"""

import os
from glob import glob
from struct import unpack
from random import choice
from itertools import product
from collections import deque, namedtuple
from twisted.internet.reactor import seconds
from twisted.internet.task import LoopingCall
from pyspades.contained import BlockAction, SetColor
from pyspades.common import make_color
from pyspades.constants import BUILD_BLOCK
from piqueserver.commands import command, player_only
from piqueserver.config import config

grownade_section = config.section("grownade")
FLYING_MODELS = grownade_section.option("flying_models")
GROW_ON_WATER = grownade_section.option("grow_on_water")

KV6_DIR = "./"+config.config_dir+'/kv6'
LOWEST_Z = 63 if GROW_ON_WATER.get() else 62
GROW_INTERVAL = 0.3

S_NO_KV6_FOLDER = "You haven't created a kv6 folder yet!"
S_NO_MATCHES = "Couldn't find any model files matching {expression}"
S_LOADED_SINGLE = 'Loaded model {filename}'
S_LOADED = 'Loaded {filenames}. Each grenade will pick one at random'
S_FAILED = 'Failed to load {filenames}'
S_PIVOT_TIP = 'Make sure the pivot points are correctly placed'
S_CANCEL = 'No longer spawning models'
S_SPECIFY_FILES = 'Specify model files to load. Wildcards are allowed, ' \
    'e.g.: "bunker", "tree*"'


class ModelLoadFailure(Exception):
    pass


def load_models(expression):
    if not os.path.isdir(KV6_DIR):
        raise ModelLoadFailure(S_NO_KV6_FOLDER)
    if not os.path.splitext(expression)[-1]:
        # append default extension
        expression += '.kv6'
    paths = glob(os.path.join(KV6_DIR, expression))
    if not paths:
        raise ModelLoadFailure(S_NO_MATCHES.format(expression=expression))

    # attempt to load models, discard invalid ones
    models, loaded, failed = [], [], []
    for path in paths:
        model = KV6Model(path)
        filename = os.path.split(path)[-1]
        if model.voxels:
            models.append(model)
            loaded.append(filename)
        else:
            failed.append(filename)
    return models, loaded, failed


@command('model', admin_only=True)
@player_only
def model_grenades(connection, expression=None):
    protocol = connection.protocol
    player = connection

    result = None
    if expression:
        try:
            models, loaded, failed = load_models(expression)
        except ModelLoadFailure as err:
            result = str(err)
        else:
            if len(loaded) == 1:
                result = S_LOADED_SINGLE.format(filename=loaded[0])
            elif len(loaded) > 1:
                result = S_LOADED.format(filenames=', '.join(loaded))

            if failed:
                player.send_chat(S_FAILED.format(filenames=', '.join(failed)))
                player.send_chat(S_PIVOT_TIP)

            if models:
                player.grenade_models = models

    elif player.grenade_models:
        player.grenade_models = None
        result = S_CANCEL
    else:
        result = S_SPECIFY_FILES
    if result:
        player.send_chat(result)


KV6Voxel = namedtuple('KV6Voxel', 'b g r a z neighbors normal_index')
NonSurfaceKV6Voxel = KV6Voxel(40, 64, 103, 128, 0, 0, 0)


class KV6Model:
    """
    Custom implementation that also generates non-surface voxels.
    Not suitable for general purpose.
    """

    size = None
    pivot = None
    voxels = None

    def __init__(self, path):
        with open(path, 'rb') as file:
            file.read(4)  # 'Kvxl'

            self.size = unpack('III', file.read(4 * 3))
            self.pivot = tuple(int(n) for n in unpack('fff', file.read(4 * 3)))

            voxel_count, = unpack('I', file.read(4))
            voxels = []
            for i in range(voxel_count):
                voxel = KV6Voxel._make(unpack('BBBBHBB', file.read(8)))
                voxels.append(voxel)

            size_x, size_y, size_z = self.size
            file.read(4 * size_x)  # discard extra information

            voxel_map = {}
            voxel_iter = iter(voxels)
            for x, y in product(range(size_x), range(size_y)):
                last_z = None
                column_len, = unpack('H', file.read(2))
                for i in range(column_len):
                    voxel = next(voxel_iter)
                    voxel_map[(x, y, voxel.z)] = voxel
                    if last_z is not None and not voxel.neighbors & 0b00010000:
                        for z in range(last_z + 1, voxel.z):
                            inside_voxel = NonSurfaceKV6Voxel._replace(z=z)
                            voxel_map[(x, y, z)] = inside_voxel
                    last_z = voxel.z

            if self.pivot not in voxel_map:
                return
            self.voxels = voxel_map


class BuildQueue:
    interval = 0.02
    blocks_per_cycle = 3
    blocks = None
    loop = None
    call_on_exhaustion = None

    def __init__(self, protocol, call_on_exhaustion=None):
        self.protocol = protocol
        self.blocks = deque()
        self.loop = LoopingCall(self.cycle)
        self.loop.start(self.interval)
        self.call_on_exhaustion = call_on_exhaustion

    def cycle(self):
        if not self.blocks:
            self.loop.stop()
            if self.call_on_exhaustion:
                self.call_on_exhaustion()
            return
        blocks_left = self.blocks_per_cycle
        last_color = None
        while self.blocks and blocks_left:
            x, y, z, color = self.blocks.popleft()
            if color != last_color:
                set_color = SetColor()
                set_color.value = make_color(*color)
                set_color.player_id = 32
                self.protocol.broadcast_contained(set_color, save=True)
                last_color = color
            if not self.protocol.map.get_solid(x, y, z):
                block_action = BlockAction()
                block_action.value = BUILD_BLOCK
                block_action.player_id = 32
                block_action.x = x
                block_action.y = y
                block_action.z = z
                self.protocol.broadcast_contained(block_action, save=True)
                self.protocol.map.set_point(x, y, z, color)
                blocks_left -= 1
        self.protocol.update_entities()

    def push_block(self, x, y, z, color):
        if not self.loop.running:
            self.loop.start(self.interval)
        self.blocks.append((x, y, z, color))


class GrowModel:
    model = None
    x, y, z = None, None, None
    open, closed = None, None
    build_queue = None
    grow_loop = None

    def __init__(self, protocol, model, x, y, z):
        self.protocol = protocol
        self.model = model
        self.x, self.y, self.z = x, y, z
        self.open = [model.pivot]
        self.closed = set()
        self.build_queue = BuildQueue(protocol, self.queue_exhausted)
        self.grow_loop = LoopingCall(self.grow_cycle)
        self.grow_loop.start(GROW_INTERVAL)

    def grow_cycle(self):
        new_nodes = set()
        for xyz in self.open:
            if xyz not in self.model.voxels or xyz in self.closed:
                continue
            self.closed.add(xyz)
            x, y, z = xyz
            new_nodes.add((x, y, z - 1))
            new_nodes.add((x, y, z + 1))
            new_nodes.add((x, y - 1, z))
            new_nodes.add((x, y + 1, z))
            new_nodes.add((x - 1, y, z))
            new_nodes.add((x + 1, y, z))
            p_x, p_y, p_z = self.model.pivot
            x, y, z = x + self.x - p_x, y + self.y - p_y, z + self.z - p_z
            if x < 0 or y < 0 or z < 0 or x >= 512 or y >= 512 or z >= LOWEST_Z:
                continue
            voxel = self.model.voxels[xyz]
            self.build_queue.push_block(x, y, z, (voxel.r, voxel.g, voxel.b))
        self.open = new_nodes
        if not new_nodes:
            self.grow_loop.stop()

    def queue_exhausted(self):
        if not self.open:
            self.release()

    def release(self):
        if self.build_queue.loop.running:
            self.build_queue.loop.stop()
        self.build_queue.loop = None
        if self.grow_loop.running:
            self.grow_loop.stop()
        self.grow_loop = None
        self.protocol.growers.remove(self)


def apply_script(protocol, connection, config):
    class SeedyConnection(connection):
        grenade_models = None

        def on_reset(self):
            self.grenade_models = None
            connection.on_reset(self)

        def on_grenade_thrown(self, grenade):
            if self.grenade_models:
                grenade.name = 'seed'
                grenade.callback = self.seed_exploded
            connection.on_grenade_thrown(self, grenade)

        def seed_exploded(self, grenade):
            if not self.grenade_models:
                return
            x, y, z = (int(n) for n in grenade.position.get())
            map = self.protocol.map
            if (not FLYING_MODELS.get() and not map.get_solid(x, y, z + 1) and
                    not z == LOWEST_Z == 63):
                return
            model = choice(self.grenade_models)
            grower = GrowModel(self.protocol, model, x, y, z)
            self.protocol.growers.append(grower)

    class SeedyProtocol(protocol):
        growers = None

        def on_map_change(self, map):
            self.growers = []
            protocol.on_map_change(self, map)

        def on_map_leave(self):
            for grower in self.growers[:]:
                grower.release()
            self.growers = None
            protocol.on_map_leave(self)

    return SeedyProtocol, SeedyConnection



"""
Runs the garbage collector at a given interval and displays any uncollected
garbage found.

.. codeauthor:: mat^2
"""

from twisted.internet import reactor
from twisted.internet.task import LoopingCall
import gc

INTERVAL = 60 * 10
VERBOSE = False


def apply_script(protocol, connection, config):
    def run_gc():
        ret = gc.collect()
        if VERBOSE:
            print('gc.collect() ->', ret)
        if not gc.garbage:
            return
        print('Memory leak detected!')
        print('Contents of gc.garbage:', gc.garbage)
    loop = LoopingCall(run_gc)
    loop.start(INTERVAL, now=False)
    return protocol, connection



from piqueserver.commands import command, get_player, PermissionDenied, player_only, target_player

@command("client", "cli")
@target_player
def client(connection, player):
    """
    Tell you information about your client or the client of a given player
    /client [player]
    """
    if connection is player:
        who_is = "You are"
    else:
        who_is = player.name + " is"

    return "{} connected with {}".format(who_is, player.client_string)


@command()
def weapon(connection, value):
    """
    Tell you what weapon a given player is using
    /weapon <player>
    """
    player = get_player(connection.protocol, value)
    if player.weapon_object is None:
        name = '(unknown)'
    else:
        name = player.weapon_object.name
    return '%s has a %s' % (player.name, name)


@command()
@player_only
def intel(connection):
    """
    Inform you of who has the enemy intel
    /intel
    """
    flag = connection.team.other.flag
    if flag.player is not None:
        if flag.player is connection:
            return "You have the enemy intel, return to base!"
        else:
            return "%s has the enemy intel!" % flag.player.name
    return "Nobody in your team has the enemy intel"


@command()
def kill(connection, value=None):
    """
    Kill yourself or a given player
    /kill [target]
    """
    if value is None:
        player = connection
    else:
        if not connection.rights.kill and not connection.admin:
            raise PermissionDenied("you can't kill other players")
        player = get_player(connection.protocol, value, False)
    player.kill()
    if connection is not player:
        message = '%s killed %s' % (connection.name, player.name)
        connection.protocol.send_chat(message, irc=True)


@command(admin_only=True)
@target_player
def heal(connection, player):
    """
    Heal and refill yourself or a given player and inform everyone on the server of this action
    /heal [player]
    """
    if connection is player:
        message = '%s was healed' % (connection.name)
    else:
        message = '%s was healed by %s' % (player.name, connection.name)
    player.refill()
    connection.protocol.send_chat(message, irc=True)


@command()
def deaf(connection, value=None):
    """
    Make you or a given player no longer receive chat messages
    /deaf [player]
    """
    if value is not None:
        if not connection.admin and not connection.rights.deaf:
            return 'No administrator rights!'
        connection = get_player(connection.protocol, value)
    message = '%s deaf' % ('now' if not connection.deaf else 'no longer')
    connection.protocol.irc_say('%s is %s' % (connection.name, message))
    message = "You're " + message
    if connection.deaf:
        connection.deaf = False
        connection.send_chat(message)
    else:
        connection.send_chat(message)
        connection.deaf = True



"""
Minefield map extension.

Allows mappers to specify the map bounds, outside of which players will trip
mines.  Breaking blocks (when standing close to the field) also triggers a
mine.

.. warning::
   This script conflicts with `smartnade` script.

Commands
^^^^^^^^

* ``/minedebug`` toggles debug mode *admin only*

example extension from mapname.txt:

>>> extensions = {
...     'minefields' : [
...         #this minefield defines the border:
...         {
...             'border' : 1,
...             'left' : 59,
...             'top' : 154,
...             'right' : 451,
...             'bottom' : 355,
...         },
...         # this specifies an additional minefield (and shows a simpler
...         # syntax)
...         {
...             'area' : (183, 126, 224, 233),    #top left
...             'height' : 60 # this specifies until which block mines are
...                           # enabled (so you can build over it)
...         }
...     ]
... }

.. codeauthor:: learn_more (MIT LICENSE)
"""
# todo: reset intel in minefield

MINEFIELD_VERSION = 1.6

from pyspades.world import Grenade
from pyspades.contained import GrenadePacket, BlockAction, SetColor
from pyspades.common import Vertex3, make_color
from pyspades.collision import collision_3d
from pyspades.constants import DESTROY_BLOCK, SPADE_DESTROY, BUILD_BLOCK
from twisted.internet.reactor import callLater
from piqueserver.commands import command, admin
from random import choice

KILL_MESSAGES = [
    '{player} wandered into a minefield',
    '{player} should not walk into a minefield',
    '{player} was not carefull enough in the minefield',
    '{player} thought those mines were toys!',
    # JoJoe's messages
    '{player} showed the team the minefield. What a hero!',
    '{player} should not use their spade to defuse mines.',
    '{player} made a huge mess in the minefield.',
    '{player} detected a mine.',
    '{player} has concluded the mine-sweeping demonstration.',
    '{player} now marks the edge of the minefield.',
    '{player} disarmed a mine by stomping on it.',
    '{player} should build a bridge across the minefield next time.',
    '{player} forgot about the minefield.',
    'Ummmm {player}, Accidentally , the minefield...',
    '{player} made soup in the minefield.',
    'Score Minefield: {mine_kills}, {player} 0, Minefield is Winning!!'
]


MINEFIELD_TIP = 'Be careful, there are mines in this map!'
MINEFIELD_MOTD = 'Watch ot for minefields!'
MINEFIELD_HELP = 'There are mines in this map!'
MINEFIELD_DBG_MESSAGE = 'Block is at x={x}, y={y}, z={z}, in field:"{m}"'
MINEFIELD_TYPE_STRING = '{type} field({left}, {top}, {right}, {bottom})'
MINEFIELD_MINE_ENT = 'mine'


class Minefield:

    def __init__(self, ext):
        self.isBorder = ext.get('border', False)
        area = ext.get('area', False)
        if area:
            self.left, self.top, self.right, self.bottom = area
        else:
            self.left = ext.get('left', 0)
            self.top = ext.get('top', 0)
            self.right = ext.get('right', 512)
            self.bottom = ext.get('bottom', 512)
        self.height = ext.get('height', 0)

    def __str__(self):
        type = 'Border' if self.isBorder else 'Inner'
        return MINEFIELD_TYPE_STRING.format(
            type=type,
            left=self.left,
            top=self.top,
            right=self.right,
            bottom=self.bottom)

    def isValid(self):
        return self.left < self.right and self.top < self.bottom

    def check_hit(self, x, y, z):
        if z >= self.height:
            if self.isBorder:
                return self.left >= x or self.right <= x or self.top >= y or self.bottom <= y
            return x >= self.left and x <= self.right and y >= self.top and y <= self.bottom
        return False

    def singleBlock(self, protocol, x, y, z, color):
        if not protocol.map.get_solid(x, y, z):
            z += 1
        if protocol.map.get_color(x, y, z) == color:
            return
        block_action = BlockAction()
        block_action.x = x
        block_action.y = y
        block_action.z = z
        block_action.player_id = 32
        protocol.map.set_point(x, y, z, color)
        block_action.value = DESTROY_BLOCK
        protocol.broadcast_contained(block_action, save=True)
        block_action.value = BUILD_BLOCK
        protocol.broadcast_contained(block_action, save=True)

    def updateColor(self, protocol, color):
        set_color = SetColor()
        set_color.value = make_color(*color)
        set_color.player_id = 32
        protocol.broadcast_contained(set_color, save=True)
        return color

    def spawnDecal(self, connection, x, y, z):
        protocol = connection.protocol
        c = self.updateColor(protocol, (0, 0, 0))
        self.singleBlock(protocol, x, y, z, c)
        c = self.updateColor(protocol, (50, 50, 50))
        for cc in ((0, 1), (1, 0), (-1, 0), (0, -1)):
            self.singleBlock(protocol, x + cc[0], y + cc[1], z, c)
        c = self.updateColor(protocol, (100, 100, 100))
        for cc in ((1, 1), (1, -1), (-1, 1), (-1, -1)):
            self.singleBlock(protocol, x + cc[0], y + cc[1], z, c)

    def spawnNade(self, connection, x, y, z):
        protocol = connection.protocol
        fuse = 0.1
        position = Vertex3(x, y, z)
        orientation = None
        velocity = Vertex3(0, 0, 0)
        grenade = protocol.world.create_object(
            Grenade,
            fuse,
            position,
            orientation,
            velocity,
            connection.grenade_exploded)
        grenade.name = MINEFIELD_MINE_ENT
        grenade_packet = GrenadePacket()
        grenade_packet.value = grenade.fuse
        grenade_packet.player_id = 32
        grenade_packet.position = position.get()
        grenade_packet.velocity = velocity.get()
        protocol.broadcast_contained(grenade_packet)
        if z >= 61.5:
            callLater(fuse + 0.1, self.spawnDecal, connection, x, y, z)


def parseField(ext):
    m = Minefield(ext)
    if m.isValid():
        return m
    return None


@command(admin_only=True)
def minedebug(connection):
    proto = connection.protocol
    proto.minefield_debug = not proto.minefield_debug
    message = 'Minefield is now in debug' if proto.minefield_debug else 'Minefield is no longer in debug'
    proto.send_chat(message, global_message=True)
    return 'You toggled minefield debug'


def apply_script(protocol, connection, config):
    class MineConnection(connection):

        def on_position_update(self):
            ret = connection.on_position_update(self)
            if self.protocol.minefield_debug:
                return ret
            pos = self.world_object.position
            x, y, z = int(pos.x), int(pos.y), int(pos.z) + 3
            if self.world_object.crouch:
                z -= 1
            self.protocol.check_mine(self, x, y, z, spawnUp=True)
            return ret

        def on_block_destroy(self, x, y, z, mode):
            if self.protocol.minefield_debug:
                message = MINEFIELD_DBG_MESSAGE.format(
                    x=x, y=y, z=z, m=self.protocol.minefieldAt(
                        x, y, z) or 'None')
                self.send_chat(message)
                return False
            if mode == DESTROY_BLOCK or mode == SPADE_DESTROY:
                pos = self.world_object.position
                #xx, yy, zz = x + 0.5, y + 0.5, z + 0.5
                if collision_3d(x, y, z, pos.x, pos.y, pos.z, 10):
                    self.protocol.check_mine(self, x, y, z)
            return connection.on_block_destroy(self, x, y, z, mode)

        def on_kill(self, killer, type, grenade):
            if grenade and grenade.name == MINEFIELD_MINE_ENT:
                self.protocol.mine_kills += 1
                message = choice(KILL_MESSAGES).format(
                    player=self.name, mine_kills=self.protocol.mine_kills)
                self.protocol.send_chat(message, global_message=True)
            connection.on_kill(self, killer, type, grenade)

    class MineProtocol(protocol):
        minefield_enabled = False
        minefields = []
        minefield_version = MINEFIELD_VERSION
        minefield_debug = False

        def on_map_change(self, map):
            self.minefields = []
            self.minefield_enabled = False
            self.mine_kills = 0
            self.minefield_debug = False
            extensions = self.map_info.extensions
            for f in extensions.get('minefields', []):
                m = parseField(f)
                if m is not None:
                    self.minefields.append(m)
            self.minefield_enabled = len(self.minefields) > 0
            return protocol.on_map_change(self, map)

        def addif(self, lst, entry):
            if lst is None or entry is None:
                return
            if self.minefield_enabled:
                if entry not in lst:
                    lst.append(entry)
            else:
                if entry in lst:
                    lst.remove(entry)

        def update_format(self):
            protocol.update_format(self)
            self.addif(self.tips, MINEFIELD_TIP)
            self.addif(self.motd, MINEFIELD_MOTD)
            self.addif(self.help, MINEFIELD_HELP)

        def minefieldAt(self, x, y, z):
            if self.minefield_enabled:
                for m in self.minefields:
                    if m.check_hit(x, y, z):
                        return m
            return None

        def check_mine(self, connection, x, y, z, waitTime=0.1, spawnUp=False):
            m = self.minefieldAt(x, y, z)
            if m:
                if spawnUp:
                    z -= 1
                callLater(waitTime, m.spawnNade, connection,
                          x + 0.5, y + 0.5, z + 0.5)

    return MineProtocol, MineConnection



# Copyright (c) James Hofmann 2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from .vxl cimport VXLData, MapData

cdef extern from "classicgen_c.cpp":
    void genland(unsigned long seed, MapData * mapdata)

import array
import random
import math
import sys
from collections import deque
cimport cython

def generate_classic(seed):
    cdef VXLData mapdata = VXLData()
    genland(seed, mapdata.map)
    return mapdata

class Biome(object):
    def __init__(self, gradient, height, variation, noise):
        """
        Create a biome with a Gradient object, typical height(0.0-1.0), and
        height variation(0.0-1.0).
        """
        self.gradient = gradient
        self.height = height
        self.variation = variation
        self.noise = noise
        self.id = -1

@cython.final
cdef class BiomeMap:
    """
    A tilemap containing biome data for a HeightMap.
    """
    cdef public int width
    cdef public int height
    cdef public int twidth
    cdef public int theight
    cdef public list biomes
    cdef public list tmap
    cdef public list gradients

    def __init__(self, biomes, width=32, height=32):
        self.biomes = biomes
        self.width = width
        self.height = height
        self.twidth = 512//self.width
        self.theight = 512//self.height
        self.tmap = [biomes[0] for n in range(self.width*self.height)]
        self.gradients = []

        for n in range(len(biomes)):
            biomes[n].id = n
            self.gradients.append(biomes[n].gradient)

    cpdef inline object get_repeat(self, int x, int y):
        """This allows the algorithm to tile at the edges."""
        return self.tmap[(x%self.width)+(y%self.height)*self.width]

    cpdef inline set_repeat(self, int x, int y, object val):
        """This allows the algorithm to tile at the edges."""
        self.tmap[(x%self.width)+(y%self.height)*self.width] = val

    cpdef noise(self):
        for idx in range(len(self.tmap)):
            self.tmap[idx] = random.choice(self.biomes)

    cpdef random_points(self, qty, biome, x=0, y=0, w=None,
                       h=None):
        """Generate some points for point_flood()"""
        result = []
        if w is None:
            w = self.width
        if h is None:
            h = self.height
        for n in range(qty):
            result.append((random.randint(x,x+w),
                         random.randint(y,y+h),
                         biome))
        return result

    cpdef point_flood(self, points):
        """Each tuple of (x,y,biome) in the "points" list
        is round-robined through a flooding
        algorithm. The algorithm uses one queue for each flood,
        so that the flooding is as even as possible."""

        openp = deque([deque([p]) for p in points])
        closedp = set([])
        biomeid = 0

        while len(openp)>0:
            plist = openp.popleft()
            p = plist.popleft()
            closedp.add((p[0],p[1]))
            self.set_repeat(p[0],p[1],p[2])
            if p[0]>0 and (p[0]-1,p[1]) not in closedp:
                plist.append((p[0]-1,p[1],p[2]))
            if p[0]<self.width-1 and (p[0]+1,p[1]) not in closedp:
                plist.append((p[0]+1,p[1],p[2]))
            if p[1]>0 and (p[0],p[1]-1) not in closedp:
                plist.append((p[0],p[1]-1,p[2]))
            if p[1]<self.height-1 and (p[0],p[1]+1) not in closedp:
                plist.append((p[0],p[1]+1,p[2]))
            if len(plist)>0:
                openp.append(plist)

    cpdef jitter(self):
        cdef int x
        cdef int y
        for idx in range(len(self.tmap)):
            x = idx % self.width
            y = idx // self.height
            self.tmap[idx] = self.get_repeat(x + random.randint(-1,1),
                                             y + random.randint(-1,1))
    cpdef create_heightmap(self):
        """Return a HeightMap with unfinished color data and a list of
        gradients. When finished with post-processing, use
        hmap.rewrite_gradient_fill(gradients). """
        cdef HeightMap hmap = HeightMap(0.)

        # paste a rectangle into each biome's area

        for idx in range(len(self.tmap)):
            x = idx % self.width
            y = idx // self.height
            biome = self.tmap[idx]
            hmap.rect_noise(x*self.twidth,y*self.theight,
                            self.twidth,self.theight,
                            biome.noise,
                            biome.height+random.random()*biome.variation)
            hmap.rect_color(x*self.twidth,y*self.theight,
                            self.twidth,self.theight,
                            biome.id)

        return hmap, self.gradients

    cpdef rect_of_point(self, x, y):
        x_pos = x*self.twidth
        y_pos = y*self.theight
        return [x_pos, y_pos, x_pos+self.twidth, y_pos+self.theight]

@cython.final
cdef class HeightMap:
    cdef public int width
    cdef public int height
    cdef public object hmap
    cdef public object cmap

    def __init__(self, height):
        self.width = 512
        self.height = 512
        self.hmap = array.array('f',[height for n in range(0,
                                    self.width*self.height)])
        self.cmap = array.array('i',[<int>0xFF00FFFF for n in range(0,
                                    self.width*self.height)])

    cpdef inline double get(self, int x, int y):
        return self.hmap[x+y*self.height]

    cpdef inline double get_repeat(self, int x, int y):
        """This allows the algorithm to tile at the edges."""
        return self.hmap[(x%self.width)+(y%self.height)*self.width]

    cpdef inline set(self, int x, int y, double val):
        self.hmap[x+y*self.height] = val

    cpdef inline set_repeat(self, int x, int y, double val):
        """This allows the algorithm to tile at the edges."""
        self.hmap[(x%self.width)+(y%self.height)*self.width] = val

    cpdef inline add_repeat(self, int x, int y, double val):
        self.hmap[(x%self.width)+(y%self.height)*self.width] += val

    cpdef inline int get_col(self, int x, int y):
        return self.cmap[x+y*self.height]

    cpdef inline int get_col_repeat(self, int x, int y):
        return self.cmap[(x%self.width)+(y%self.height)*self.width]

    cpdef inline set_col_repeat(self, int x, int y, int val):
        self.cmap[(x%self.width)+(y%self.height)*self.width] = val

    cpdef inline fill_col(self, int col):
        for n in range(len(self.cmap)):
            self.cmap[n] = col

    cpdef mult_repeat(self, int x, int y, double mult):
        cdef int idx = (x%self.width)+(y%self.height)*self.width
        self.hmap[idx] *= mult

    cpdef seed(self, double jitter, double midpoint):
        cdef double halfjitter = jitter * 0.5
        for idx in range(len(self.hmap)):
            self.hmap[idx] = midpoint + (random.random()*jitter -
                                         halfjitter)

    cpdef peaking(self):
        """Adds a "peaking" feel to the map."""
        for idx in range(len(self.hmap)):
            self.hmap[idx] = self.hmap[idx] * self.hmap[idx]

    cpdef dipping(self):
        """Adds a "dipping" feel to the map."""
        for idx in range(len(self.hmap)):
            self.hmap[idx] = math.sin(self.hmap[idx]*(math.pi))

    cpdef rolling(self):
        """Adds a "rolling" feel to the map."""
        for idx in range(len(self.hmap)):
            self.hmap[idx] = math.sin(self.hmap[idx]*(math.pi/2))

    cpdef smoothing(self):
        """Does some simple averaging to bring down the noise level."""
        for x in range(0,self.width):
            for y in range(0,self.height):
                top = self.get_repeat(x,y-1)
                left = self.get_repeat(x-1,y)
                right = self.get_repeat(x+1,y)
                bot = self.get_repeat(x,y+1)
                center = self.hmap[x+y*self.width]
                self.hmap[x+y*self.width] = (top + left + right + bot + center)/5

    cpdef midpoint_displace(self, double jittervalue, \
                          double spanscalingmultiplier, \
                            int skip=0):
        """Midpoint displacement with the diamond-square algorithm."""

        cdef int span = self.width+1
        cdef float spanscaling = 1.
        cdef float jitterrange
        cdef float jitteroffset
        cdef int halfspan
        cdef float topleft
        cdef float topright
        cdef float botleft
        cdef float botright
        cdef float center

        for iterations in range(9): # hardcoded for 512x512
            if skip>0:
                skip-=1
                span = span >> 1
                spanscaling = spanscaling * spanscalingmultiplier
                continue
            jitterrange = jittervalue * spanscaling
            jitteroffset = - jitterrange / 2
            for x in range(0,self.width,span):
                for y in range(0,self.height,span):
                    halfspan = span >> 1
                    topleft = self.get_repeat(x,y)
                    topright = self.get_repeat((x+span),y)
                    botleft = self.get_repeat(x,(y+span))
                    botright = self.get_repeat((x+span),(y+span))
                    center = (topleft+topright+botleft+botright) * 0.25\
                             + (random.random() * jitterrange + jitteroffset)

                    self.set_repeat(x+halfspan,y,(topleft+topright+center)*0.33)
                    self.set_repeat(x,y+halfspan,(topleft+botleft+center)*0.33)
                    self.set_repeat(x+halfspan,y+span,
                                    (botleft+botright+center)*0.33)
                    self.set_repeat(x+span,y+halfspan,
                                    (topright+botright+center)*0.33)
                    self.set_repeat(x + halfspan, y + halfspan, center)
            span = span >> 1
            spanscaling = spanscaling * spanscalingmultiplier

    cpdef jitter_heights(self, double amount):
        """Image jittering filter. Amount is max pixels distance to jitter."""
        cdef int nx = 0
        cdef int ny = 0
        cdef int idx = 0

        while idx<len(self.hmap):
            nx = int((idx % self.width) + (random.random()-0.5)*amount)
            ny = int((idx // self.width) + (random.random()-0.5)*amount)
            self.hmap[idx] = self.get_repeat(nx, ny)
            idx+=1

    cpdef jitter_colors(self, double amount):
        """Image jittering filter. Amount is max pixels distance to jitter."""
        cdef int nx = 0
        cdef int ny = 0
        cdef int idx = 0

        while idx<len(self.hmap):
            nx = int((idx % self.width) + (random.random()-0.5)*amount)
            ny = int((idx // self.width) + (random.random()-0.5)*amount)
            self.cmap[idx] = self.get_col_repeat(nx, ny)
            idx+=1

    cpdef level_against_heightmap(self, HeightMap other, double height):
        """Use another HeightMap as an alpha-mask to force values to a
            specific height"""
        for x in range(0, self.width):
            for y in range(0, self.height):
                orig = self.get_repeat(x,y)
                dist = orig - height
                self.set_repeat(x,y, orig - dist * other.get_repeat(x,y))

    cpdef blend_heightmaps(self, HeightMap alphamap, HeightMap HeightMap):
        """Blend according to two HeightMaps: one as an alpha-mask,
            the other contains desired heights"""
        for x in range(0, self.width):
            for y in range(0, self.height):
                orig = self.get_repeat(x,y)
                dist = orig - HeightMap.get_repeat(x,y)
                self.set_repeat(x,y, orig - dist * alphamap.get_repeat(x,y))

    cpdef rect_solid(self, int x, int y, int w, int h, double z):
        for xx in range(x, x+w):
            for yy in range(y, y+h):
                self.set_repeat(xx,yy,z)

    cpdef rect_noise(self, int x, int y, int w, int h,
                     double jitter, double midpoint):
        cdef double halfjitter = jitter * 0.5
        for xx in range(x,x+w):
            for yy in range(y,y+h):
                self.set(xx,yy, midpoint + (random.random()*jitter -
                                         halfjitter))
    cpdef rect_color(self, int x, int y, int w, int h, int col):
        for xx in range(x, x+w):
            for yy in range(y, y+h):
                self.set_col_repeat(xx,yy,col)

    cpdef truncate(self):
        """Truncates the HeightMap to a valid (0-1) range.
        Do this before painting or writing to voxels to avoid crashing."""
        for idx in range(0,len(self.hmap)):
            self.hmap[idx] = min(max(self.hmap[idx],0.0),1.0)

    cpdef offset_z(self, double qty):
        for idx in range(0,len(self.hmap)):
            self.hmap[idx] = self.hmap[idx]+qty

    cpdef rescale_z(self, double multiple):
        for idx in range(0,len(self.hmap)):
            self.hmap[idx] = self.hmap[idx]*multiple

    cpdef paint_gradient_fill(self, gradient):
        """Surface the map with a single gradient."""
        cdef zcoldef = gradient.array()

        cdef int idx = 0

        while idx<len(self.hmap):
            h = int(self.hmap[idx] * 63)
            self.cmap[idx] = paint_gradient(zcoldef,h)
            idx+=1

    cpdef rewrite_gradient_fill(self, list gradients):
        """Given a cmap of int-indexed gradient definitions,
        rewrite them as surface color definitions."""

        cdef zcoldefs = []
        for n in gradients:
            zcoldefs.append(n.array())

        cdef int idx = 0

        while idx<len(self.hmap):
            h = int(self.hmap[idx] * 63)
            self.cmap[idx] = paint_gradient(zcoldefs[self.cmap[idx]],h)
            idx+=1

    cpdef rgb_noise_colors(self, low, high):
        """Add noise to the heightmap colors."""
        cdef int idx = 0

        patterns = array.array('i', [random.randint(low,high) for n in range(101)])

        while idx<len(self.hmap):
            mid = self.cmap[idx]

            r = max(0, min(0xFF,get_r(mid)+patterns[idx%len(patterns)]))
            g = max(0, min(0xFF,get_g(mid)+patterns[(idx+1)%len(patterns)]))
            b = max(0, min(0xFF,get_b(mid)+patterns[(idx+2)%len(patterns)]))

            self.cmap[idx] = make_color(r,g,b)

            idx+=1

    cpdef smooth_colors(self):
        """Average the color of each pixel to add smoothness."""
        cdef int x = 0
        cdef int y = 0

        import copy

        swap = copy.deepcopy(self.cmap)

        while y<self.height:
            left = swap[((x-1)%self.width)+(y%self.height)*self.width]
            right = swap[((x+1)%self.width)+(y%self.height)*self.width]
            up = swap[((x)%self.width)+((y-1)%self.height)*self.width]
            down = swap[((x)%self.width)+((y+1)%self.height)*self.width]
            mid = swap[((x)%self.width)+((y)%self.height)*self.width]

            r = (get_r(left) + get_r(right) + get_r(up) + get_r(down) + get_r(mid))//5
            g = (get_g(left) + get_g(right) + get_g(up) + get_g(down) + get_g(mid))//5
            b = (get_b(left) + get_b(right) + get_b(up) + get_b(down) + get_b(mid))//5

            self.set_col_repeat(x,y,make_color(r,g,b))

            x += 1
            if x>=self.width:
                x = 0
                y += 1

    cpdef write_vxl(self):
        cdef VXLData vxl = VXLData()

        cdef int x = 0
        cdef int y = 0
        cdef int h = 0
        cdef int z = 0
        cdef int idx = 0

        while idx<len(self.hmap):
            x = idx % self.width
            y = idx // self.height
            h = int(self.hmap[idx] * 63)
            vxl.set_column_fast(x, y, h, 63, int(min(63,h+3)),
                                self.cmap[idx])
            idx+=1
        return vxl

    cpdef line_add(self,int x,int y,
                int x2,int y2,int radius, double depth):
        cdef int posx, posy
        for c in bresenham_line(x,y,x2,y2):
            posx = c[0]
            posy = c[1]
            for x in range(-radius,radius+1):
                for y in range(-radius,radius+1):
                    self.add_repeat(posx+x,posy+y,depth)

    cpdef line_set(self,int x,int y,
                int x2,int y2,int radius, double height):
        cdef int posx, posy
        for c in bresenham_line(x,y,x2,y2):
            posx = c[0]
            posy = c[1]
            for x in range(-radius,radius+1):
                for y in range(-radius,radius+1):
                    self.set_repeat(posx+x,posy+y,height)


cdef lim_byte(int val):
    return max(0,min(255,val))


cpdef inline int make_color(int r, int g, int b):
    return b | (g << 8) | (r << 16) | (<int>128 << 24)


cpdef inline int get_r(int color):
    return (color>>16) & 0xFF


cpdef inline int get_g(int color):
    return (color>>8) & 0xFF


cpdef inline int get_b(int color):
    return (color) & 0xFF


cdef inline int paint_gradient(object zcoltable, int z):
    cdef int zz = z*3
    cdef int rnd = random.randint(-4,4)
    return make_color(lim_byte(zcoltable[zz]+rnd),
                      lim_byte(zcoltable[zz+1]+rnd),
                      lim_byte(zcoltable[zz+2]+rnd)
                      )


cdef inline list bresenham_line(int x, int y, int x2, int y2):
    cdef int steep = 0
    cdef list coords = []
    cdef int dx, dy, sx, sy, d
    dx = abs(x2 - x)
    if (x2 - x) > 0: sx = 1
    else: sx = -1
    dy = abs(y2 - y)
    if (y2 - y) > 0: sy = 1
    else: sy = -1
    if dy > dx:
        steep = 1
        x,y = y,x
        dx,dy = dy,dx
        sx,sy = sy,sx
    d = (2 * dy) - dx
    for i in range(0,dx):
        if steep: coords.append((y,x))
        else: coords.append((x,y))
        while d >= 0:
            y = y + sy
            d = d - (2 * dx)
        x = x + sx
        d = d + (2 * dy)
    coords.append((x2,y2))
    return coords


from pyspades.color import hsb_to_rgb, interpolate_hsb, interpolate_rgb


class Gradient(object):
    def __init__(self):
        self.steps = []
        for n in range(0,64):
            self.steps.append((0,0,0,0))

    def set_step_rgb(self, step, rgb):
        self.steps[step] = (rgb[0],rgb[1],rgb[2],255)

    def set_step_hsb(self, step, hsb):
        rgb = hsb_to_rgb(*hsb)
        self.steps[step] = (rgb[0],rgb[1],rgb[2],255)

    def rgb(self, start_pos, start_color, end_pos, end_color):
        """Linear interpolation of (0-255) RGB values."""
        dist = end_pos - start_pos
        for n in range(start_pos, end_pos):
            pct = float(n - start_pos) / dist
            self.set_step_rgb(n, interpolate_rgb(start_color,
                                             end_color,
                                             pct))

    def hsb(self, start_pos, start_color, end_pos, end_color):
        """Linear interpolation of (0-360,0-100,0-100) HSB values
            as used in GIMP."""
        dist = end_pos - start_pos
        start_color = (start_color[0]/360.,
                       start_color[1]/100.,
                       start_color[2]/100.)
        end_color = (end_color[0]/360.,
                       end_color[1]/100.,
                       end_color[2]/100.)
        for n in range(start_pos, end_pos):
            pct = float(n - start_pos) / dist
            interp = interpolate_hsb(start_color, end_color, pct)
            self.set_step_hsb(n, interp)

    def array(self):
        base = list(self.steps)
        base.reverse()
        result = []
        for rgb in base:
            result.append(int(rgb[0]))
            result.append(int(rgb[1]))
            result.append(int(rgb[2]))
        return array.array('i',result)
